"""
The :mod:`py_fatigue.msc.corrections` module collects all the
mean stress corrections implemented in py_fatigue.
classes.
"""

from typing import Union

import matplotlib.pyplot as plt
import numpy as np

from ..utils import make_axes


def dnvgl_mean_stress_correction(
    mean_stress: np.ndarray,
    stress_amplitude: np.ndarray,
    detail_factor: Union[float, int] = 0.8,
    plot: bool = False,
) -> np.ndarray:
    """Calculates the mean stress correction according to par.
    2.5 of`DNVGL-RP-C203 <https://bit.ly/3dUZ1OY>`_ which includes
    an attenuation factor :math:`p` for the stress ranges if the
    following cases:

    * base material without significant residual stresses \
    :math:`\\rightarrow p = 0.6`. This option neglects fully \
    compressive cycles.
    * welded connections without significant residual stresses \
    :math:`\\rightarrow p = 0.8`. This option multiplies the \
    stress range of fully compressive cycles by p.

    Given that the stress ranges are :math:`\\Delta \\sigma`,
    the corrected stress ranges are:

    .. math::

        \\Delta \\sigma_{corr} = f_m \\cdot \\Delta \\sigma

    where:

    .. math::

        f_m = \\frac{\\sigma_{max} + p \\cdot \\vert \\sigma_{min}
        \\vert}{\\sigma_{max} + \\vert\\sigma_{min} \\vert}

    with:

    .. math::

        \\frac{\\sigma_a}{\\vert \\sigma_m \\vert} \\leq 1, \\quad
        \\sigma_a = \\frac{\\sigma_{max} - \\sigma_{min}}{2} \\,\\land
        \\, \\sigma_m = \\frac{\\sigma_{max} + \\sigma_{min}}{2}

    Parameters
    ----------
    mean_stress : np.ndarray
        The mean stress of the fatigue test (:math:`\\sigma_m`).
    stress_amplitude : np.ndarray
        The stress amplitude of the fatigue test (:math:`\\sigma_a`).
    detail_factor : float, optional
        mean stress attenuation factor (:math:`p`), defaults to 0.8
        in welded connections without significant residual stresses.
    plot : bool, optional
        If True, the mean stress correction is plotted, defaults to
        False.

    Returns
    -------
    np.ndarray
        The mean stress corrected ranges
        (:math:`\\Delta \\sigma_{corr}`).
    """

    max_stress = mean_stress + stress_amplitude
    min_stress = mean_stress - stress_amplitude

    f_m = (max_stress + detail_factor * np.abs(min_stress)) / (
        max_stress + np.abs(min_stress)
    )

    f_m[f_m > 1] = 1
    f_m[min_stress > 0] = 1
    if detail_factor == 0.8:
        f_m[f_m < detail_factor] = detail_factor
    elif detail_factor == 0.6:
        f_m[f_m < detail_factor] = 0
    else:
        raise ValueError(
            f"Detail factor = {str(detail_factor)} is not allowed in "
            + "DNVGL-RP-C203. Only 0.6 and 0.8 are permitted."
        )

    stress_range_corr = 2 * f_m * stress_amplitude

    # NOTE: next plot has been included for testing purposes
    if plot:  # pragma: no cover
        # from matplotlib.ticker import MaxNLocator
        _, axes = make_axes()
        xp, yp = zip(*sorted(zip(min_stress / max_stress, f_m)))
        axes.plot(
            xp,
            yp,
            linestyle=":",
            linewidth=1,
            color="#000000",
            marker=".",
            markersize=6,
            markerfacecolor="#000000",
            markeredgewidth=0.5,
            markeredgecolor="#666666",
        )
        plt.rcParams.update({"mathtext.default": "regular"})

        axes.set_xlabel("load ratio, $\\sigma_{min}/ \\sigma_{max}$")
        axes.set_ylabel("$f_m$")
        axes.spines["right"].set_visible(False)
        axes.spines["top"].set_visible(False)
        axes.set_xlim([-20, 20])

    return stress_range_corr


def walker_mean_stress_correction(
    mean_stress: np.ndarray,
    stress_amplitude: np.ndarray,
    gamma: float = 0.5,
    plot: bool = False,
) -> np.ndarray:
    """Calculates the mean stress correction according to Walker model.

    The correction is given by:

    .. math::

        \\Delta \\sigma_{corr} = {\\sigma_{max}} ^ {(1 - \\gamma)} \\,
        \\sigma_{alt} ^ {\\gamma}

    with:

    .. math::

        \\sigma_{max} = \\sigma_{mean} + \\sigma_{alt}

    Parameters
    ----------
    mean_stress : np.ndarray
        The mean stress of the fatigue test (:math:`\\sigma_{mean}`).
    stress_amplitude : np.ndarray
        The stress amplitude of the fatigue test (:math:`\\sigma_{alt}`).
    gamma : float, optional
        The gamma Walker exponent (:math:`\\gamma`), defaults to 0.5.
    plot : bool, optional
        If True, the mean stress correction is plotted, defaults to
        False.

    Returns
    -------
    np.ndarray
        The mean stress corrected ranges (:math:`\\Delta \\sigma_{corr}`).
    """

    max_stress = mean_stress + stress_amplitude

    stress_range_corr = np.nan_to_num(
        (max_stress ** (1 - gamma)) * (stress_amplitude**gamma)
    )

    # NOTE: next plot has been included for testing purposes
    if plot:  # pragma: no cover
        # from matplotlib.ticker import MaxNLocator
        _, axes = make_axes()
        xp, yp = zip(*sorted(zip(2 * stress_amplitude, stress_range_corr)))
        axes.plot(
            xp,
            yp,
            linestyle=":",
            linewidth=1,
            color="#000000",
            marker=".",
            markersize=6,
            markerfacecolor="#000000",
            markeredgewidth=0.5,
            markeredgecolor="#666666",
        )
        plt.rcParams.update({"mathtext.default": "regular"})

        axes.set_xlabel("uncorrected stress range, $\\Delta \\sigma$")
        axes.set_ylabel("MSC stress range, , $\\Delta \\sigma_{MSC}$")
        axes.spines["right"].set_visible(False)
        axes.spines["top"].set_visible(False)

    return stress_range_corr


def swt_mean_stress_correction(
    mean_stress: np.ndarray,
    stress_amplitude: np.ndarray,
    plot: bool = False,
) -> np.ndarray:
    """Calculates the mean stress correction according to
    Smith-Watson-Topper model. It is equivalent to the Walker model
    with gamma = 0.5.

    The correction is given by:

    .. math::

        \\Delta \\sigma_{corr} = \\sqrt{\\sigma_{max} \\, \\sigma_{alt}}

    with:

    .. math::

        \\sigma_{max} = \\sigma_{mean} + \\sigma_{alt}

    Parameters
    ----------
    mean_stress : np.ndarray
        The mean stress of the fatigue test (:math:`\\sigma_{mean}`).
    stress_amplitude : np.ndarray
        The stress amplitude of the fatigue test (:math:`\\sigma_{alt}`).
    plot : bool, optional
        If True, the mean stress correction is plotted, defaults to
        False.

    Returns
    -------
    np.ndarray
        The mean stress corrected ranges (:math:`\\Delta \\sigma_{corr}`).
    """

    return walker_mean_stress_correction(
        mean_stress, stress_amplitude, gamma=0.5, plot=plot
    )



from .stress_life import (
    calc_dem,
    calc_des,
    calc_pm,
    calc_nonlinear_damage,
    get_dem,
    get_des,
    get_pm,
    get_nonlinear_damage,
)

from .crack_growth import get_crack_growth, CalcCrackGrowth

__all__ = [
    "calc_dem",
    "calc_des",
    "calc_pm",
    "get_dem",
    "get_des",
    "get_pm",
    "calc_nonlinear_damage",
    "get_nonlinear_damage",
    "get_crack_growth",
    "CalcCrackGrowth",
]



# -*- coding: utf-8 -*-

"""The :mod:`py_fatigue.material.paris_curve` module collects all
functions related to :term:`Paris' law` definition.
The main class is ParisCurve.
"""

# Standard "from" imports
from collections.abc import Collection
from typing import Any, List, Optional, Tuple, Union

# Standard imports
import abc

import io

# import itertools
# import warnings

# Non-standard "from" imports
# from plotly.express import colors as px_colors

# Non-standard imports
import matplotlib.figure
import matplotlib.axis
import matplotlib.pyplot as plt
import numba as nb
import numpy as np

# import plotly.graph_objs as go
# import plotly.io as pio

# py_fatigue imports
from ..utils import make_axes, calc_slope_intercept
from ..material.sn_curve import (
    _set_color,
    _get_name,
    _check_param_couple_types,
    ensure_array,
)


class AbstractCrackGrowthCurve(metaclass=abc.ABCMeta):
    """Abstract Paris curve. It also implements mean stress effect according
    to Walker's model.
    Concrete subclasses should define methods:

        _ `_repr_svg_`
        - `format_name`
        - `get_knee_sif`
        - `get_knee_growth_rate`
        - `get_sif`
        - `get_growth_rate`
    """

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments
    # Eight is reasonable in this case.

    def __init__(
        self,
        slope: Union[int, float, list, np.ndarray],
        intercept: Union[int, float, list, np.ndarray],
        threshold: Union[int, float] = 0,
        critical: Union[int, float] = np.inf,
        load_ratio: Union[int, float] = 0,
        walker_exponent: Union[int, float] = 0,
        environment: Optional[str] = None,
        curve: Optional[str] = None,
        norm: Optional[str] = None,
        unit_string: str = "MPa √mm",
        color: Union[str, None] = None,
    ) -> None:
        """Define crack growth rate curve (Paris' law).
        See class docstring for more information.

        Parameters
        ----------
        slope : Union[int, float, list, np.ndarray]
            Paris curve slope
        intercept : Union[int, float, list, np.ndarray]
            crack growth rate axis intercept
        threshold : Union[int, float]
            propagation threshold, below which crack growth rate is null
        critical : Union[int, float]
            critical propagation stress intensity factor, imminent failure
        load_ratio : Union[int, float]
            load ratio. Used for mean stress-corrected Paris' law
        walker_exponent : Union[int, float]
            Walker's exponent. Used for mean stress-corrected Paris' law
        environment : str, optional
            Paris' law envirnoment, by default None
        curve : str, optional
            Paris' law category, by default None
        norm : str, optional
            Paris' law norm, by default None
        unit_string : str, optional
            units, by default "MPa"
        color : str, optional
            RGBS or HEX string for color, by default None
        """
        self.color, self.id = _set_color(color)
        self.norm = norm
        self.environment = environment
        self.curve = curve
        self.name = _get_name(curve, environment, norm)
        self.__slope, self.__intercept = _check_param_couple_types(
            slope, intercept
        )
        self.__linear = len(self.slope) == 1
        self.__threshold = threshold
        self.__critical = critical
        self.__load_ratio = load_ratio
        self.__walker_exponent = walker_exponent
        self.__unit = unit_string

    @classmethod
    @abc.abstractmethod
    def from_knee_points(
        cls,
        knee_sif: Union[np.ndarray, List[float]],
        knee_growth_rate: Union[np.ndarray, List[float]],
        environment: Optional[str] = None,
        curve: Optional[str] = None,
        norm: Optional[str] = None,
        unit_string: str = "MPa √mm",
        color: Union[str, None] = None,
    ) -> "AbstractCrackGrowthCurve":
        """Create a Paris curve from a set of knee points, i.e. build slopes
        and intercepts from pairs of SIF and crack growth rate. The first
        and last knee points are used to set the threshold and critical
        SIF values."""

    def __str__(self) -> str:
        """Crack growth curve __str__ method

        Returns:
            str: Crack growth curve string description
        """
        return self.name

    def __eq__(self, other: object) -> bool:
        """Compare different crack growth curves and assert that they
        contain same attributes values
        """
        if self.__class__ != other.__class__:
            return False

        def compare_exact(first, second):
            """Return whether two dicts of arrays are exactly equal"""
            if first.keys() != second.keys():
                return False
            return all(
                np.array_equal(first[key], second[key]) for key in first
            )

        return compare_exact(self.__dict__, other.__dict__)

    @property
    def linear(self) -> bool:
        """Preventing attribute modification outside of constructor

        Returns:
            bool: Is it a linear Paris' law?
        """
        return self.__linear

    @property
    def slope(self) -> np.ndarray:
        """Preventing attribute modification outside of constructor

        Returns:
            np.ndarray: Slope
        """
        return self.__slope

    @property
    def intercept(self) -> np.ndarray:
        """Preventing attribute modification outside of constructor

        Returns:
            np.ndarray: Intercept
        """
        return self.__intercept

    @property
    def threshold(self) -> Union[int, float]:
        """Preventing attribute modification outside of constructor

        Returns:
            Union[int, float]: Threshold SIF
        """
        return self.__threshold

    @property
    def critical(self) -> Union[int, float]:
        """Preventing attribute modification outside of constructor

        Returns:
            Union[int, float]: Critical SIF
        """
        return self.__critical

    @property
    def load_ratio(self) -> Union[int, float]:
        """Preventing attribute modification outside of constructor

        Returns:
            Union[int, float]: Load ratio
        """
        return self.__load_ratio

    @property
    def walker_exponent(self) -> Union[int, float]:
        """Preventing attribute modification outside of constructor

        Returns:
            Union[int, float]: Walker's exponent
        """
        return self.__walker_exponent

    @property
    def w_intcpt(self) -> np.ndarray:
        """Walker-corrected intercept.

        .. math::
            a_w = \\frac{a}{(1 - R)^{\\frac{m}{1 - \\gamma}}}

        Preventing attribute modification outside of constructor

        Returns:
            np.ndarray: Corrected intercept
        """
        return self.__intercept * (1 - self.__load_ratio) ** (
            self.__slope * (self.__walker_exponent - 1)
        )

    @property
    def unit(self) -> str:
        """Preventing attribute modification outside of constructor

        Returns:
            unit: Units of measure, by default "MPa"
        """
        return self.__unit

    @abc.abstractmethod
    def format_name(self, html_format: bool = False) -> None:
        """Reformat ParisCurve name.

        Parameters
        ----------
        html_format : bool, optional
            Choose whether the name format shall be HTML or not,
            by default True
        """

    @property
    @abc.abstractmethod
    def threshold_growth_rate(self) -> float:
        """Calculates the crack growth rate at threshold, if threshold
        SIF is defined.

        Returns
        -------
        float
            threshold crack growth rate
        """

    @property
    @abc.abstractmethod
    def critical_growth_rate(self) -> float:
        """Calculates the crack growth rate at critical, if critical
        SIF is defined.

        Returns
        -------
        float
            critical crack growth rate
        """

    @abc.abstractmethod
    def get_knee_growth_rate(
        self,
        check_knee: Optional[Collection] = None,
        significant_digits: int = 2,
    ) -> np.ndarray:
        """Calculates the crack growth rate at the knee, if the Paris'
        law is more than linear.

        Parameters
        ----------
        check_knee : Collection, optional
            Iterable of SIF values to check for the knee, by default None
        significant_digits : int, optional
            Number of significant digits to round the knee to, by default 2

        Returns
        -------
        1darray
            knee crack growth rate
        """

    @abc.abstractmethod
    def get_knee_sif(
        self,
        check_knee: Optional[Collection] = None,
        significant_digits: int = 2,
    ) -> np.ndarray:
        """Calculates the SIF at the knee, if the Paris' law is more
        than linear.

        Parameters
        ----------
        check_knee : iterable, optional
            SIF values to check the knee SIF against, by default None
        significant_digits : int, optional
            number of significant digits to check the knee SIF against,
            by default 2

        Returns
        -------
        np.ndarray
            knee SIF
        """

    @abc.abstractmethod
    def get_growth_rate(
        self, sif_range: Union[int, float, list, np.ndarray]
    ) -> Union[float, np.ndarray]:
        """Return cycles value(s) for stress range(s).

        Parameters
        ---------
        sif_range: int or float or list or 1darray
           :term:`SIF range(s)<SIF range>` to find the corresponding
           :term:`crack growth rate(s)<Crack growth rate>` for.

        Returns
        -------
        float or 1darray
            :term:`crack growth rate(s)<Crack growth rate>` for the
            :term:`SIF range(s)<SIF range>` indicating `sif_range`.
        """

    @abc.abstractmethod
    def get_sif(
        self, growth_rate: Union[int, float, list, np.ndarray]
    ) -> Union[float, np.ndarray]:
        """Return :term:`SIF range(s)<SIF range>` for the endurance(s) N.

        Parameters
        ---------
        cycles : int or float or list or 1darray
            :term:`crack growth rate(s)<Crack growth rate>` to find the
            corresponding :term:`SIF range(s)<SIF range>` for.

        Returns
        -------
        float or 1darray
            :term:`SIF range(s)<SIF range>` for corresponding
            :term:`crack growth rate(s)<Crack growth rate>`.
        """


class ParisCurve(AbstractCrackGrowthCurve):
    """Define a crack growth curve inthe form of Paris' law that can
    have an arbitrary number of slopes/intercepts as well as a
    threshold and critical stress intensity factor. For example:

    >>> #
    >>> #         ^                log da/dN - log ΔK
    >>> #         │                                *
    >>> #         │                                *
    >>> #         │                                *
    >>> #         │                                *
    >>> #         │                             *  .
    >>> #         │                          *  │  .
    >>> #         │                       *     │ m2
    >>> #         │                    *        │  .
    >>> #         │                 *-----------┘  .
    >>> #         │              *                 .
    >>> #         │           .*│                  .
    >>> #         │        . *  │ m1               .
    >>> #         │     .  *    │                  .
    >>> #         │  .   *------┘                  .
    >>> # (da/dN)1+    *                           .
    >>> #         │  . *                           .
    >>> #         │.   *                           .
    >>> # (da/dN)2+----|---------------------------|--------------->
    >>> #             ΔK0                         ΔKc
    >>> #                          Number of cycles

    The slope-intercept couples `((m1, log_a1), (m2, log_a2))`, and the
    endurance value `(Ne)` are the parameters necessary to fully describe this
    trilinear Paris' law. If the endurance is not set, it defaults to Inf.

    .. note::
        The curve can account for mean stress effect according to Walker's
        model. The walker_exponent is set to 0 by default, meaning no
        correction is applied, i.e. the base curve is valid for R=0.

    Example
    -------

    >>> from py_fatigue import ParisCurve
    >>> import numpy as np

    >>> # Define a Paris' law with 5 slopes and 5 interceps
    >>> SIF = np.linspace(1,2500, 300)
    >>> SLOPE_5 = np.array([2.88, 5.1, 8.16, 5.1, 2.88])
    >>> INTERCEPT_5 = np.array([1E-16, 1E-20, 1E-27, 1E-19, 1E-13])
    >>> THRESHOLD = 20
    >>> CRITICAL = 2000

    >>> pc = ParisCurve(slope=SLOPE_5, intercept=INTERCEPT_5,
                      threshold=THRESHOLD, critical=CRITICAL, norm="The norm",
                      environment="Environment", curve="nr. 4")
    >>> pc_.get_knee_sif()
    array([ 63.35804993, 193.9017369 , 411.50876026, 504.31594872])

    """

    @classmethod
    def from_knee_points(
        cls,
        knee_sif: Union[np.ndarray, List[float]],
        knee_growth_rate: Union[np.ndarray, List[float]],
        environment: Optional[str] = None,
        curve: Optional[str] = None,
        norm: Optional[str] = None,
        unit_string: str = "MPa √mm",
        color: Union[str, None] = None,
    ) -> "ParisCurve":
        """Create a Paris curve from a set of knee points, i.e. build slopes
        and intercepts from pairs of SIF and crack growth rate. The first
        and last knee points are used to set the threshold and critical
        SIF values.

        Parameters
        ----------
        knee_sif : Union[np.ndarray, List[float]]
            SIF values at the knee
        knee_growth_rate : Union[np.ndarray, List[float]]
            crack growth rate at the knee
        environment : str, optional
            Paris' law environment, by default None
        curve : str, optional
            Paris' law category, by default None
        norm : str, optional
            Paris' law norm, by default None
        unit_string : str, optional
            units, by default "MPa"
        color : str, optional
            RGBS or HEX string for color, by default None

        Returns
        -------
        ParisCurve
            Paris curve instance
        """
        # We first need to prepend the threshold (x,y) and append the critical
        # (x, y) pairs
        _s, _i = calc_slope_intercept(
            np.log(knee_sif), np.log(knee_growth_rate)
        )

        return cls(
            slope=_s,
            intercept=np.exp(_i),
            threshold=np.min(knee_sif),
            critical=np.max(knee_sif),
            environment=environment,
            curve=curve,
            norm=norm,
            unit_string=unit_string,
            color=color,
        )

    def _repr_svg_(self) -> str:  # pragma: no cover
        fig, ax = self.plot()
        # fig.dpi = 100
        fig.set_figheight(5)
        fig.set_figwidth(7)
        max_x = np.max(ax.lines[0].get_xdata())
        max_y = np.max(ax.lines[0].get_ydata())
        cg_unit = "[L][CYCLE]$^{-1}$"
        if "√" in self.unit:
            split_unit = self.unit.split("√")
            cg_unit = "/".join((split_unit[1], "cycle"))

        ann_1 = f"{self.name},\nSlope: {[round(s, 2) for s in self.slope]}"
        ann_2 = f"Intercept: {[f'{i:.2E}' for i in self.w_intcpt]}".replace(
            "'", ""
        )
        ann_3 = f"Threshold: {self.threshold:.2f} {cg_unit}"
        ann_4 = f"Critical: {self.critical:.2f}, {self.unit}"
        annotation = "\n".join((ann_1, ann_2, ann_3, ann_4))
        ax.text(
            1.5 * max_x,
            max_y,
            annotation,
            horizontalalignment="left",
            verticalalignment="top",
            color="#435580",
            bbox={
                "facecolor": "#FFF",
                "alpha": 0.95,
                "pad": 10,
                "edgecolor": "#0E76BC",
            },
            fontsize=8,
            fontweight="semibold",
        )
        if not self.linear:
            knee_zip = zip(self.get_knee_sif(), self.get_knee_growth_rate())
            for j, (knee_sif, knee_gr) in enumerate(knee_zip):
                annotation = (
                    f"Knee {j +1}: ({round(knee_sif, 2)}, {knee_gr:.2E})"
                )
                ax.plot(knee_sif, knee_gr, "o", color="#C40000")
                for item in (
                    [ax.title, ax.xaxis.label, ax.yaxis.label]
                    + ax.get_xticklabels()
                    + ax.get_yticklabels()
                ):
                    item.set_fontsize(10)
                ax.text(
                    1.1 * knee_sif,
                    self.get_growth_rate(knee_sif),
                    annotation,
                    horizontalalignment="left",  # if j % 2 == 0 else "left",
                    verticalalignment="center",
                    fontsize=8,
                    color="#C40000",
                )
        if self.critical < np.inf:
            annotation = (
                f"Critical: ({round(self.critical, 2)}, "
                + f"{self.critical_growth_rate:.2E})"
            )
            ax.plot(
                self.critical, self.critical_growth_rate, "d", color="#435580"
            )
            ax.text(
                0.9 * self.critical,
                self.critical_growth_rate,
                annotation,
                horizontalalignment="right",
                verticalalignment="center",
                fontsize=8,
                color="#435580",
            )
        if self.threshold > 0:
            annotation = (
                f"Threshold: ({round(self.threshold, 2)}, "
                + f"{self.threshold_growth_rate:.2E})"
            )
            ax.plot(
                self.threshold,
                self.threshold_growth_rate,
                "d",
                color="#435580",
            )
            ax.text(
                1.1 * self.threshold,
                self.threshold_growth_rate,
                annotation,
                horizontalalignment="left",
                verticalalignment="center",
                fontsize=8,
                color="#435580",
            )
        fig.tight_layout()

        output = io.BytesIO()
        fig.savefig(output, format="svg")
        data = output.getvalue()  # .encode('utf-8') doesn't change anything
        plt.close(fig)
        return data.decode("utf-8")

    def format_name(self, html_format: bool = False) -> None:
        if not html_format:
            self.name = self.name.replace("<br> ", "\n")
        else:
            self.name = self.name.replace("\n", "<br> ")

    @property
    def threshold_growth_rate(self) -> float:
        """Calculates the crack growth rate at threshold, if threshold
        SIF is defined.

        Returns
        -------
        float
            threshold crack growth rate
        """
        return self.get_growth_rate(self.threshold)

    @property
    def critical_growth_rate(self) -> float:
        """Calculates the crack growth rate at critical, if critical
        SIF is defined.

        Returns
        -------
        float
            critical crack growth rate
        """
        return self.get_growth_rate(self.critical)

    def get_knee_growth_rate(
        self,
        check_knee: Optional[Collection] = None,
        significant_digits: int = 2,
    ) -> np.ndarray:
        """Calculates the crack growth rate at the knee, if the Paris'
        law is more than linear.

        Parameters
        ----------
        check_knee : Iterable, optional
            Iterable of SIF values to check for the knee, by default None
        significant_digits : int, optional
            Number of significant digits to round the knee to, by default 2

        Returns
        -------
        np.ndarray
            knee crack growth rate
        """
        knee_growth_rate = np.empty(self.w_intcpt.size - 1, dtype=np.float64)
        if not self.linear:
            for i in range(self.w_intcpt.size - 1):
                m_i = self.slope[i + 1] / (self.slope[i + 1] - self.slope[i])
                m_i_p_1 = self.slope[i] / (self.slope[i + 1] - self.slope[i])
                knee_growth_rate[i] = (
                    self.w_intcpt[i] ** m_i / self.w_intcpt[i + 1] ** m_i_p_1
                )
            if check_knee is not None:
                try:
                    iter(check_knee)
                except TypeError:
                    check_knee = np.asarray([check_knee])
                else:
                    check_knee = np.asarray(check_knee)
                if len(check_knee) != len(self.slope) - 1:  # type: ignore
                    raise ValueError(
                        f"{len(self.slope) - 1} knee points expected"
                    )
                check_knee_growth_rate = self.get_growth_rate(check_knee)
                _ = [
                    np.testing.assert_approx_equal(  # type: ignore
                        c_k_s, k_s, significant=significant_digits
                    )
                    for c_k_s, k_s in zip(
                        check_knee_growth_rate, knee_growth_rate
                    )
                ]
        else:
            if check_knee is not None:
                raise ValueError("0 knee points expected")
        return knee_growth_rate

    def get_knee_sif(
        self,
        check_knee: Optional[Collection] = None,
        significant_digits: int = 2,
    ) -> np.ndarray:
        """Calculates the SIF at the knee, if the Paris' law is more
        than linear.

        Parameters
        ----------
        check_knee : iterable, optional
            SIF values to check the knee SIF against, by default None
        significant_digits : int, optional
            number of significant digits to check the knee SIF against,
            by default 2

        Returns
        -------
        np.ndarray
            knee SIF
        """
        knee_sif = np.empty(self.w_intcpt.size - 1, dtype=np.float64)
        if not self.linear:
            if self.w_intcpt.size > 1:
                for i in range(self.w_intcpt.size - 1):
                    knee_sif[i] = (
                        self.w_intcpt[i] / self.w_intcpt[i + 1]
                    ) ** (1 / (self.slope[i + 1] - self.slope[i]))
            if check_knee is not None:
                try:
                    len(check_knee)
                except TypeError:
                    check_knee = np.asarray([check_knee])
                else:
                    check_knee = np.asarray(check_knee)
                if len(check_knee) != len(self.slope) - 1:  # type: ignore
                    raise ValueError(
                        f"{len(self.slope) - 1} knee points expected"
                    )
                _ = [
                    np.testing.assert_approx_equal(  # type: ignore
                        c_k_s, k_s, significant=significant_digits
                    )
                    for c_k_s, k_s in zip(check_knee, knee_sif)  # type: ignore
                ]
        else:
            if check_knee is not None:
                raise ValueError("0 knee points expected")
        return knee_sif

    @ensure_array
    def get_growth_rate(
        self, sif_range: Union[int, float, list, np.ndarray]
    ) -> Union[float, np.ndarray]:
        return _calc_growth_rate(
            sif_range,
            self.slope,
            self.w_intcpt,
            self.threshold,
            self.critical,
        )

    @ensure_array
    def get_sif(
        self, growth_rate: Union[int, float, list, np.ndarray]
    ) -> Union[float, np.ndarray]:
        return _calc_sif(
            growth_rate,
            self.slope,
            self.w_intcpt,
            self.threshold,
            self.critical,
        )

    def plot(
        self,
        sif: Optional[list] = None,
        growth_rate: Optional[list] = None,
        dataset_name: Optional[str] = None,
        dataset_color: str = "#000",
        fig: Optional[matplotlib.figure.Figure] = None,
        ax: Optional[matplotlib.axes.Axes] = None,  # type: ignore
        **kwargs: Any,
    ) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:  # type: ignore
        """Use matplotlib to plot the Paris' law and a da/dN vs ΔK
        history dataset.

        Example
        -------

        Use matplotlib to plot the Paris' law and a da/dN vs ΔK history

        >>> fig, ax = pc.plot()

        Parameters
        ----------
        sif : list, optional
            SIF value(s) in the dataset,
            by default None
        growth_rate : list, optional
            :term:`Crack growth rate(s)<Crack growth rate>` in the dataset,
            by default None
        dataset_name : str, optional
            history dataset name, by default None
        dataset_color : str, optional
            history dataset color, by default "#000"
        fig : matplotlib.figure.Figure, optional
            figure object, by default None
        ax : matplotlib.axes.Axes, optional
            axis object, by default None
        **kwargs : Any, optional
            additional keyword arguments

        Returns
        -------
        matplotlib.figure.Figure, matplotlib.axes.Axes
            The figure and axes.
        """
        fig, ax = make_axes(fig, ax)
        sif_plot, growth_rate_plot = _paris_curve_data_points(self)

        # plt.figure(figsize=(6, 2.5))
        ax.loglog(  # type: ignore
            sif_plot,
            growth_rate_plot,
            label=self.name,
            color=self.color,
            **kwargs,
        )
        if sif is not None and growth_rate is not None:
            sif, growth_rate = _check_param_couple_types(sif, growth_rate)
            ax.loglog(  # type: ignore
                sif,
                growth_rate,
                label=dataset_name,
                color=dataset_color,
                linewidth=1,
                marker="o",
                markersize=4,
            )

        plt.grid(which="both", linestyle=":")
        ax.set_xlabel(f"SIF range (ΔK), {self.unit}")  # type: ignore
        if "√" in self.unit:
            split_unit = self.unit.split("√")
            y_unit = "/".join((split_unit[1], "cycle"))
        else:
            y_unit = "[L][CYCLE]$^{-1}$"
        ax.set_ylabel(f"Crack growth rate (da/dN), {y_unit}")  # type: ignore

        return fig, ax


@nb.njit(
    # 'float64[::1](float64[::1], float64[::1], float64[::1])',
    fastmath=False,
    parallel=True,
)
def _calc_growth_rate(sif, slope, intercept, threshold, critical):
    # pylint: disable=not-an-iterable
    assert intercept.size > 0 and intercept.size == slope.size
    assert np.min(sif) >= 0
    assert 0 <= threshold < critical <= np.inf

    knees_sif = np.empty(intercept.size - 1, dtype=np.float64)
    if intercept.size > 1:
        for i in nb.prange(intercept.size - 1):
            knees_sif[i] = (intercept[i] / intercept[i + 1]) ** (
                1 / (slope[i + 1] - slope[i])
            )
    knees_sif = np.hstack(
        (
            np.array([0.9999999999 * threshold]),
            knees_sif,
            np.array([critical / 0.9999999999]),
        )
    )
    e_msg = (
        "Knee(s) not in between threshold and critical SIF."
        + "\nCheck the definitions of slope, intercept, threshold, critical."
    )
    assert np.all(np.diff(knees_sif) > 0), e_msg
    # print("knees_sif:", knees_sif)
    idx = np.digitize(sif, knees_sif, right=False) - 1
    # print("idx:", idx)
    the_growth_rate = np.empty(sif.size, dtype=np.float64)
    for i in nb.prange(sif.size):
        if sif[i] < threshold:  # below threshold
            the_growth_rate[i] = 0
            continue
        if sif[i] > critical:  # above threshold
            the_growth_rate[i] = np.inf
            continue
        the_growth_rate[i] = intercept[idx[i]] * sif[i] ** slope[idx[i]]

    return the_growth_rate


@nb.njit(
    # 'float64[::1](float64[::1], float64[::1], float64[::1])',
    fastmath=False,
    parallel=True,
)
def _calc_sif(growth_rate, slope, intercept, threshold, critical):
    # pylint: disable=not-an-iterable
    assert intercept.size > 0 and intercept.size == slope.size
    assert np.min(growth_rate) >= 0
    assert 0 <= threshold < critical <= np.inf

    knees_growth_rate = np.empty(intercept.size - 1, dtype=np.float64)
    if intercept.size > 1:
        for i in nb.prange(intercept.size - 1):
            m_i = slope[i + 1] / (slope[i + 1] - slope[i])
            m_i_p_1 = slope[i] / (slope[i + 1] - slope[i])
            knees_growth_rate[i] = (
                intercept[i] ** m_i / intercept[i + 1] ** m_i_p_1
            )

    knees_growth_rate = np.hstack(
        (
            np.array([intercept[0] * (0.9999999999 * threshold) ** slope[0]]),
            knees_growth_rate,
            np.array([intercept[-1] * (critical / 0.9999999999) ** slope[-1]]),
        )
    )
    e_msg = (
        "Knee(s) not in between threshold and critical SIF."
        + "\nCheck the definitions of slope, intercept, threshold, critical."
    )
    assert np.all(np.diff(knees_growth_rate) > 0), e_msg
    # print("knees_growth_rate:", knees_growth_rate)
    idx = np.digitize(growth_rate, knees_growth_rate, right=False) - 1
    # print("idx:", idx)
    the_sif = np.empty(growth_rate.size, dtype=np.float64)
    nr_knees = intercept.size - 1
    for i in nb.prange(growth_rate.size):
        if idx[i] <= 0:  # below threshold
            the_sif[i] = threshold
            continue
        if idx[i] >= nr_knees:  # above threshold
            the_sif[i] = critical
            continue
        the_sif[i] = (growth_rate[i] / intercept[idx[i]]) ** (
            1 / slope[idx[i]]
        )

    return the_sif


def _paris_curve_data_points(pc: ParisCurve) -> tuple:
    """Draw Paris' curve data points, including possible knees and
    threshold/critical SIF.

    Parameters
    ----------
    pc : ParisCurve
        Instance of :class:`py_fatigue.ParisCurve`

    Returns
    -------
    tuple
        tuple of arrays
    """

    # if pc.threshold == 0 and pc.critical == np.inf:
    threshold: Union[int, float] = 1.0
    critical: Union[int, float] = 1000.0
    if pc.threshold > 0 and pc.critical < np.inf:
        threshold = pc.threshold
        critical = pc.critical
    if pc.threshold == 0 and pc.critical < np.inf:
        threshold = pc.critical / 100
        critical = pc.critical
    if pc.threshold > 0 and pc.critical == np.inf:
        threshold = pc.threshold
        critical = pc.threshold * 100

    sif_plot = np.logspace(
        np.log10(threshold), np.log10(0.999999999999 * critical), 20
    )  # 0.999999999999 to avoid rounding errors

    if not pc.linear:
        sif_plot = np.sort(np.append(sif_plot, np.array(pc.get_knee_sif())))
    growth_rate_plot = pc.get_growth_rate(sif_plot)
    if pc.threshold > 0:
        growth_rate_plot = np.hstack(
            [growth_rate_plot[0] / 10, growth_rate_plot]
        )
        sif_plot = np.hstack([pc.threshold, sif_plot])
    if pc.critical < np.inf:
        growth_rate_plot = np.hstack(
            [growth_rate_plot, growth_rate_plot[-1] * 10]
        )
        # Again, 0.999999999999 to avoid rounding errors
        sif_plot = np.hstack([sif_plot, 0.999999999999 * pc.critical])
    return sif_plot, growth_rate_plot



"""The module :mod:`py_fatigue.damage.stress_life` contains all the
damage models related to the stress-life approach.
"""

# Packages from standard library
from types import SimpleNamespace
from typing import Callable, Optional, Tuple, Union
import warnings

# Packages from external libraries
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

# py-fatigue imports
from ..cycle_count.cycle_count import CycleCount
from ..material.sn_curve import SNCurve
from ..utils import make_axes
from ..styling import py_fatigue_formatwarning
from ..material.sn_curve import _check_param_couple_types

try:
    # delete the accessor to avoid warning
    del pd.DataFrame.miner  # type: ignore
except AttributeError:
    pass


@pd.api.extensions.register_dataframe_accessor("miner")
class PalmgrenMiner:
    """Accessor for the Palmgren-Miner damage model."""

    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj
        self.sn_curve = None
        self.des_data = {}
        self.dem_data = {}

    @staticmethod
    def _validate(obj):
        """Validate the input DataFrame. Raise an error if the input
        DataFrame does not contain the right columns.
        """
        if {
            "cycles_to_failure",
        }.issubset(obj.columns):
            e_msg = "'cycles_to_failure' already calculated"
            raise AttributeError(e_msg)

        if not {"count_cycle", "mean_stress", "stress_range"}.issubset(
            obj.columns
        ):
            e_msg = (
                "Must have 'count_cycle', 'mean_stress' and 'stress_range'."
            )
            raise AttributeError(e_msg)

    def damage(self, sn_curve: SNCurve):
        """Calculate the damage of the samples.

        Parameters
        ----------
        sn_curve : SNCurve
            The SNCurve object.

        Returns
        -------
        DataFrame
            The DataFrame with the damage.
        """
        if self._obj._metadata["unit"] != sn_curve.unit:
            e_msg = (
                f"Units of cycle_count ({self._obj._metadata['unit']}) and "
                f"sn_curve ({sn_curve.unit}) do not match."
            )
            raise ValueError(e_msg)
        self._obj.sn_curve = sn_curve
        self.sn_curve = sn_curve
        self._obj["cycles_to_failure"] = sn_curve.get_cycles(
            self._obj.stress_range
        )
        self._obj["pm_damage"] = np.divide(
            self._obj.count_cycle, self._obj.cycles_to_failure
        )
        return self._obj

    def des(self, slope: float, equivalent_cycles: float = 1e7) -> float:
        """Calculate the damage equivalent stress range (DES).

        Parameters
        ----------
        slope : float
            The SN curve slope.
        equivalent_cycles : float, optional
            The equivalent number of cycles, by default 1e7

        Returns
        -------
        float
            The DES.

        See Also
        --------
        calc_des
        """
        self.des_data["slope"] = slope
        self.des_data["equivalent_cycles"] = equivalent_cycles

        return calc_des(
            self._obj.stress_range,
            self._obj.count_cycle,
            slope,
            equivalent_cycles,
        )

    def dem(
        self,
        outer_radius: float,
        inner_radius: float,
        slope: float,
        equivalent_cycles: float = 1e7,
    ) -> float:
        """Calculate the damage equivalent stress range (DES).

        Parameters
        ----------
        slope : float
            The SN curve slope.
        equivalent_cycles : float, optional
            The equivalent number of cycles, by default 1e7

        Returns
        -------
        float
            The DES.

        See Also
        --------
        calc_des
        """
        self.dem_data["outer_radius"] = outer_radius
        self.dem_data["inner_radius"] = inner_radius
        self.dem_data["slope"] = slope
        self.dem_data["equivalent_cycles"] = equivalent_cycles

        return calc_dem(
            outer_radius,
            inner_radius,
            self._obj.stress_range,
            self._obj.count_cycle,
            slope,
            equivalent_cycles,
        )

    def plot_histogram(
        self,
        fig: Optional[matplotlib.figure.Figure] = None,
        ax: Optional[matplotlib.collections.PathCollection] = None,
        dens_func: Optional[Callable] = None,
        **kwargs,
    ) -> Tuple[
        matplotlib.figure.Figure, matplotlib.collections.PathCollection
    ]:
        """Plot the damage.

        Parameters
        ----------
        fig : matplotlib.figure.Figure, optional
            The figure to plot on, by default None
        ax : matplotlib.collections.PathCollection, optional
            The axes to plot on, by default None
        kwargs
            Additional keyword arguments for the plot.

        Returns
        -------
        matplotlib.figure.Figure, matplotlib.collections.PathCollection
            The figure and axes.
        """
        fig, ax = make_axes(fig, ax)
        im = ax.scatter(
            self._obj.count_cycle,
            self._obj.stress_range,
            # edgecolors="#CCC",
            # linewidths=0.5,
            c=(
                dens_func(self._obj.pm_damage)
                if dens_func is not None
                else self._obj.pm_damage
            ),
            **kwargs,
        )
        cbar_label = (
            "PM damage"
            if dens_func is None
            else "PM damage(" + str(dens_func) + ")"
        )
        ax.set_xlabel("# of cycles")
        ax.set_ylabel("Stress range, MPa")
        cbar = plt.colorbar(im, ax=ax)
        cbar.set_label(cbar_label, rotation=270)
        cbar.ax.get_yaxis().labelpad = 10
        return fig, ax


try:
    # delete the accessor to avoid warning
    del pd.DataFrame.gassner  # type: ignore
except AttributeError:
    pass


@pd.api.extensions.register_dataframe_accessor("gassner")
class Gassner:
    """Accessor for the Gassner shift factor for variable amplitude."""

    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj
        self.sn_curve = None

    @staticmethod
    def _validate(obj):
        """Validate the input DataFrame. Raise an error if the input
        DataFrame does not contain the right columns.
        """
        if {
            "shift_factor",
        }.issubset(obj.columns):
            e_msg = "'shift_factor' already calculated"
            raise AttributeError(e_msg)

        if not {"count_cycle", "mean_stress", "stress_range"}.issubset(
            obj.columns
        ):
            e_msg = (
                "Must have 'count_cycle', 'mean_stress' and 'stress_range'."
            )
            raise AttributeError(e_msg)

    def g(self, sn_curve: SNCurve):
        """Calculate the shift factor of the samples.

        The shift factor is a spectrum-dependent factor that allows
        to shift a Wohler curve and avoids having to calculate the
        damage using the :class:`~PalmgrenMiner` ruleevery time.

        Once the shift factor has been calculated, it can be applied
        to the Wohler SN curve. Then, accssing the curve through the
        maximum amplitude value of the spectrum will return the
        expected life for said spectrum.

        The shift factor for each cycle  in the spectrum is:

        .. math::

            G_j = \\frac{n_j}{N_S} \\,
            \\left({\\frac{\\sigma_{{max}_j} \\, \\sigma_{{alt}_j}}
            {\\max{\\sigma_{alt}} ^ {2}}}\\right)^{2}

        The shift factor then is:

        .. math::

            G = \\sum_{j=1}^{N_B} G_j

        Parameters
        ----------
        sn_curve : SNCurve
            The SNCurve object.

        Returns
        -------
        DataFrame
            The DataFrame with the shift factor per sample.
        """
        assert sn_curve.linear, "The SN curve must be linear"
        self._obj.sn_curve = sn_curve
        self.sn_curve = sn_curve
        stress_amp = self._obj.stress_range / 2
        max_amp = max(stress_amp)
        nu = np.divide(self._obj.count_cycle, sum(self._obj.count_cycle))
        self._obj["shift_factor"] = nu * np.power(
            np.divide(stress_amp, max_amp), self.sn_curve.slope
        )
        return self._obj


def calc_pm(
    stress_range: Union[float, pd.Series, np.ndarray],
    count_cycle: Union[float, pd.Series, np.ndarray],
    sn_curve: SNCurve,
) -> np.ndarray:
    """Calculates the damage using the Palmgren-Miner rule, i.e. the
    linear damage accumulation rule:

    .. math::

        D = \\sum_{j=1}^{N_{\\text{blocks}}} \\frac{n_j}{N_j} \\leq 1

    Parameters
    ----------
     stress_range : Union[pd.Series, np.ndarray]
        The stress range
    count_cycle : Union[pd.Series, np.ndarray]
        The number of cycles
    sn_curve : SNCurve
        The SNCurve object.

    Returns
    -------
    Union[float, np.ndarray]
        The damage according to the Palmgren-Miner rule.
    """
    stress_range, count_cycle = _check_param_couple_types(
        stress_range, count_cycle
    )
    return count_cycle / sn_curve.get_cycles(stress_range)


def calc_des(
    stress_range: Union[float, pd.Series, np.ndarray],
    count_cycle: Union[float, pd.Series, np.ndarray],
    slope: float,
    equivalent_cycles: float = 1e7,
) -> float:
    """Calculate the damage equivalent stress range (DES).

    Parameters
    ----------
    stress_range : Union[float, pd.Series, np.ndarray]
        The stress range
    count_cycle : Union[float, pd.Series, np.ndarray]
        The number of cycles
    slope : float
        The SN curve slope
    equivalent_cycles : float, optional
        The equivalent number of cycles, by default 1e7

    Returns
    -------
    float
        The DES.
    """
    stress_range, count_cycle = _check_param_couple_types(
        stress_range, count_cycle
    )
    des = np.power(
        np.sum(
            np.multiply(
                np.power(stress_range, float(slope)),
                count_cycle,
            )
            / equivalent_cycles
        ),
        1 / slope,
    )
    return des


def calc_dem(
    outer_radius: float,
    inner_radius: float,
    stress_range: Union[float, pd.Series, np.ndarray],
    count_cycle: Union[float, pd.Series, np.ndarray],
    slope: float,
    equivalent_cycles: float = 1e7,
) -> float:
    """Calculate the damage equivalent moment (DEM) in Nm.

    .. math::

        DEM = DES * I_c / r_o = DES * S

    with :math:`I_c` the area moment of inertia and :math:`r_i` the
    inner radius, or :math:`S` the Section modulus. The Damage
    Equivalent Stress Range (DES) is calculated using
    :func:`calc_des`

    Parameters
    ----------
    outer_radius : float
        The outer radius to calculate the DEM. Must be provided in m
    inner_radius : float
        The inner radius to calculate the DEM. Must be provided in m
    stress_range : Union[float, pd.Series, np.ndarray]
        The stress range
    count_cycle : Union[float, pd.Series, np.ndarray]
        The number of cycles
    slope : float
        The SN curve slope
    equivalent_cycles : float, optional
        The equivalent number of cycles, by default 1e7

    Raises
    ------
    ValueError
        If the inner radius is larger than the outer radius.

    Returns
    -------
    float
        The DEM in Nm.
    """
    if outer_radius <= inner_radius:
        raise ValueError("outer_radius must be greater than inner_radius")
    stress_range, count_cycle = _check_param_couple_types(
        stress_range, count_cycle
    )
    area_inertia = np.pi / 4 * (outer_radius**4 - inner_radius**4)
    des = calc_des(stress_range, count_cycle, slope, equivalent_cycles)

    return area_inertia * des * 1e6 / outer_radius


def get_pm(cycle_count: CycleCount, sn_curve: SNCurve) -> np.ndarray:
    """Calculates the damage using the Palmgren-Miner rule.

    Parameters
    ----------
    cycle_count : CycleCount
        The CycleCount object.
    sn_curve : SNCurve
        The SNCurve object.

    Returns
    -------
    np.ndarray
        The damage according to the Palmgren-Miner rule.
    """
    if cycle_count.unit != sn_curve.unit:
        raise ValueError(
            f"Units of cycle_count ({cycle_count.unit}) and "
            f"sn_curve ({sn_curve.unit}) do not match."
        )
    return calc_pm(cycle_count.stress_range, cycle_count.count_cycle, sn_curve)


def get_des(
    cycle_count: CycleCount, slope: float, equivalent_cycles: float = 1e7
) -> float:
    """Calculate the damage equivalent stress range (DES).

    Parameters
    ----------
    cycle_count : CycleCount
        The CycleCount object.
    slope : float
        The SN curve slope
    equivalent_cycles : float, optional
        The equivalent number of cycles, by default 1e7

    Returns
    -------
    float
        The DES.
    """
    return calc_des(
        cycle_count.stress_range,
        cycle_count.count_cycle,
        slope,
        equivalent_cycles,
    )


def get_dem(
    outer_radius: float,
    inner_radius: float,
    cycle_count: CycleCount,
    slope: float,
    equivalent_cycles: float = 1e7,
) -> float:
    """Calculate the damage equivalent moment (DEM) in Nm.

    .. math::

        DEM = DES * I_c / r_o = DES * S

    with :math:`I_c` the area moment of inertia and :math:`r_i` the
    inner radius, or :math:`S` the Section modulus. The Damage
    Equivalent Stress Range (DES) is calculated using
    :func:`calc_des`

    Parameters
    ----------
    outer_radius : float
        The outer radius to calculate the DEM. Must be provided in m
    inner_radius : float
        The inner radius to calculate the DEM. Must be provided in m
    stress_range : Union[pd.Series, np.ndarray]
        The stress range
    count_cycle : Union[pd.Series, np.ndarray]
        The number of cycles
    slope : float
        The SN curve slope
    equivalent_cycles : float, optional
        The equivalent number of cycles, by default 1e7

    Returns
    -------
    float
        The DEM in Nm.
    """
    return calc_dem(
        outer_radius,
        inner_radius,
        cycle_count.stress_range,
        cycle_count.count_cycle,
        slope,
        equivalent_cycles,
    )


def calc_manson_halford_exponents(
    cycles_to_failure: np.ndarray, exponent: float = 0.4
) -> np.ndarray:
    """Calculate the Mannson-Halford exponents :math:`e_{j, j+1}`.

    .. math::

        e{j, j+1} = \\left(\\frac{N_{j}}{N_{j+1}}\\right)^{exponent}

    Parameters
    ----------
    cycles_to_failure : np.ndarray
        The number of cycles to failure.
    exponent : float, optional
        The exponent, by default 0.4

    Returns
    -------
    np.ndarray
        The Mannson-Halford exponents.
    """
    return (
        np.hstack([cycles_to_failure[:-1] / cycles_to_failure[1:], [1]])
        ** exponent
    )


def calc_pavlou_exponents(
    stress_range: np.ndarray,
    ultimate_stress: float = 900,
    exponent: float = -0.75,
) -> np.ndarray:
    """Calculate the Pavlou exponents :math:`q(\\sigma_j)=e_{j, j+1}`.

    .. math::

        e_{j, j+1} = \\left(\\frac{\\Delta \\sigma_j / 2}{\\sigma_U}\\right)
        ^{exponent}

    where :math:`\\Delta \\sigma_j` is the stress amplitude, :math:`\\sigma_U`
    is the ultimate stress, :math:`\\Delta \\sigma` is the stress range and
    :math:`exponent` is the exponent.


    Parameters
    ----------
    stress_range : np.ndarray
        The stress range.
    ultimate_stress : float, optional
        The ultimate stress, by default 900
    exponent : float, optional
        The exponent, by default -0.75

    Returns
    -------
    np.ndarray
        The Pavlou exponents.
    """
    q_exp = (stress_range / 2 / ultimate_stress) ** exponent
    return np.hstack([q_exp[1:] / q_exp[:-1], [1]])


def calc_si_jian_exponents(
    stress_range: np.ndarray,
) -> np.ndarray:
    """Calculate the Si-Jian et al. exponents as

    .. math::

        e_{j, j+1} = \\sigma_{j+1} / \\sigma_{j}

    where :math:`\\sigma_{j+1}` is the stress amplitude for the
    :math:`j`-th cycle.

    Parameters
    ----------
    stress_range : np.ndarray
        The stress range.

    Returns
    -------
    np.ndarray
        The Si-Jian et al. exponents.
    """
    return np.hstack([stress_range[1:] / stress_range[:-1], [1]])


def calc_nonlinear_damage(
    damage_rule: str,
    stress_range: np.ndarray,
    count_cycle: np.ndarray,
    sn_curve: SNCurve,
    **kwargs,
) -> float:
    """Calculate the fatigue damage using a nonlinear damage rule
    among the allowed ones:

    - 'Pavlou': Pavlou damage rule
    - 'Manson-Halford': Mannson-Halford damage rule
    - 'Si-Jian': Si-Jian et al damage rule
    - 'Leve': Leve damage rule

    The generic form of a nonlinear damage rule is:

    .. math::

        D = \\left(
            \\left( \\dots
                \\left(
                    \\left(
                        \\left(\\frac{n_1}{N_1}\\right)^{e_{1, 2}} +
                        \\frac{n_2}{N_2}
                    \\right)^{e_{2, 3}} +
                    \\frac{n_3}{N_3}
                \\right)^{e_{3, 4}} + \\dots + \\frac{n_{M-1}}{N_{M-1}}
            \\right)^{e_{M-1, M}} + \\dots + \\frac{n_M}{N_M}
        \\right)^{e_M}

    where :math:`n_j` is the number of cycles in the fatigue histogram
    at the :math:`j`-th cycle, :math:`N_j` is the number of cycles to
    failure at the :math:`j`-th cycle, :math:`e_{j, j+1}` is the exponent
    for the :math:`j`-th and :math:`j+1`-th cycles, :math:`M` is the
    number of load blocks in the fatigue spectrum.

    The formula is conveniently rewritten as pseudocode:

    .. code-block:: python
        :caption: pseudocode for the nonlinear damage rule

        # retrieve N_j using the fatigue histogram and SN curve
        # retrieve the exponents e_{j, j+1}
        #  calculate the damage
        D = 0
        for j in range(1, M+1):
            D = (D + n_j / N_j) ^ e_{j, j+1}

    Parameters
    ----------
    damage_rule : str
        The damage rule to use. Must be one of the following:
        'Pavlou', 'Manson-Halford', 'Si-Jian', 'Leve'.
    stress_range : np.ndarray
        The stress range.
    count_cycle : np.ndarray
        The number of cycles.
    sn_curve : SNCurve
        The SN curve.
    kwargs : dict
        The keyword arguments for the damage rule. The following
        keyword arguments are allowed:

        - 'base_exponent': The exponent for the damage rule.
        - 'ultimate_stress': The ultimate stress.

    Returns
    -------
    float
        The cumulated damage.
    """
    damage_per_cycle = calc_pm(
        stress_range,
        count_cycle,
        sn_curve,
    )
    allowed_rules = [
        "pavlou",
        "manson",
        "manson and halford",
        "manson halford",
        "leve",
        "si jian",
        "si jian et al",
    ]
    if damage_rule not in allowed_rules:
        e_msg = f"damage_rule must be one of {allowed_rules}"
        raise ValueError(e_msg)

    ns = SimpleNamespace(**kwargs)
    if damage_rule.lower() == "pavlou":
        if "base_exponent" not in kwargs:
            w_msg = [
                "Pavlou damage rule requires 'base_exponent' ",
                "to be assigned.\nUsing preset value of -0.75.",
            ]
            warnings.formatwarning = py_fatigue_formatwarning
            warnings.warn("".join(w_msg), UserWarning)
            ns.base_exponent = -0.75
        if "ultimate_stress" not in kwargs:
            w_msg = [
                "Pavlou damage rule requires 'ultimate_stress' ",
                "to be assigned.\nUsing preset value of 900 MPa.",
            ]
            warnings.formatwarning = py_fatigue_formatwarning
            warnings.warn("".join(w_msg), UserWarning)
            ns.ultimate_stress = 900
        damage_exp = calc_pavlou_exponents(
            stress_range, ns.ultimate_stress, ns.base_exponent
        )
    if "manson" in damage_rule.lower():
        if "base_exponent" not in kwargs:
            w_msg = [
                "Manson & Halford damage rule requires ",
                "'base_exponent' to be assigned.\n",
                "Using preset value of 0.4.",
            ]
            warnings.formatwarning = py_fatigue_formatwarning
            warnings.warn("".join(w_msg), UserWarning)
            ns.base_exponent = 0.4
        damage_exp = calc_manson_halford_exponents(
            sn_curve.get_cycles(stress_range), ns.base_exponent
        )
    if damage_rule.lower() == "leve":
        if "base_exponent" not in kwargs:
            w_msg = [
                "Leve damage rule requires ",
                "'base_exponent' to be assigned.\n",
                "Using preset value of 2.",
            ]
            warnings.formatwarning = py_fatigue_formatwarning
            warnings.warn("".join(w_msg), UserWarning)
            ns.base_exponent = 2
        damage_exp = ns.base_exponent * np.ones(len(damage_per_cycle))
    if "si jian" in damage_rule.lower():
        damage_exp = calc_si_jian_exponents(stress_range)

    total_damage = 0
    damage_array = np.empty(len(damage_per_cycle))
    for i, (d_i, e_i) in enumerate(zip(damage_per_cycle, damage_exp)):
        total_damage = (total_damage + d_i) ** e_i
        damage_array[i] = total_damage

    return damage_array[-1]


def get_nonlinear_damage(
    damage_rule: str,
    cycle_count: CycleCount,
    sn_curve: SNCurve,
    **kwargs,
) -> float:
    """Calculate the fatigue damage using a nonlinear damage rule.

    Parameters
    ----------
    damage_rule : str
        The damage rule to use. Must be one of the following:
        'Pavlou', 'Manson-Halford', 'Si-Jian', 'Leve'.
    cycle_count : CycleCount
        The cycle count object.
    sn_curve : SNCurve
        The SN curve object.
    kwargs : dict
        The keyword arguments for the damage rule. The following
        keyword arguments are allowed:

        - 'base_exponent': The exponent for the damage rule.
        - 'ultimate_stress': The ultimate stress.

    Returns
    -------
    float
        The cumulated damage.
    """
    if cycle_count.unit != sn_curve.unit:
        e_msg = "The units of the cycle count and SN curve must be the same."
        raise ValueError(e_msg)
    return calc_nonlinear_damage(
        damage_rule,
        cycle_count.stress_range,
        cycle_count.count_cycle,
        sn_curve,
        **kwargs,
    )



"""The module :mod:`py_fatigue.damage.crack_growth` contains all the
damage models related to the crack growth approach.
"""

# Packages from the Python Standard Library
from typing import Dict, Tuple, Type, TypeVar

# Packages from third party libraries
from numba.experimental import jitclass
import numba as nb
import numpy as np
import pandas as pd

# Local packages
from ..cycle_count.cycle_count import CycleCount
from ..material.crack_growth_curve import ParisCurve
from ..utils import split, to_numba_dict
from ..geometry.generic import AbstractCrackGeometry
from ..geometry.cylinder import f_hol_cyl_01


try:
    # delete the accessor to avoid warning
    del pd.DataFrame.cg  # type: ignore
except AttributeError:
    pass

ACG = TypeVar("ACG", bound=AbstractCrackGeometry)

spec = [
    ("stress_range", nb.float64[::1]),
    ("count_cycle", nb.float64[:]),
    ("slope", nb.float64[::1]),
    ("intercept", nb.float64[::1]),
    ("threshold", nb.float64),
    ("critical", nb.float64),
    ("crack_type", nb.types.string),
    ("crack_geometry", nb.types.DictType(nb.types.unicode_type, nb.float64)),
    ("crack_depth", nb.float64[::1]),
    ("sif", nb.float64[::1]),
    ("geometry_factor", nb.float64[::1]),
    ("final_cycles", nb.float64),
    ("failure", nb.boolean),
]


@jitclass(spec)
class CalcCrackGrowth:  # pragma: no cover
    """Crack growth object. This object is used to calculate the crack
    growth rate and the crack size. The class has been implemented
    using the Numba library to speed up the calculations.

    The crack size is numerically integrated over the crack growth
    curve. In this example we will consider the Paris' law, but the
    function can be used with any other crack growth curve.

    .. math::

        \\begin{align}
        \\begin{cases}
            \\frac{da}{dN} & = C \\left( \\Delta K \\right)^{m} \\\\
            \\Delta K & = Y(a) \\cdot \\Delta \\sigma \\sqrt{\\pi a} \\\\
        \\end{cases}
        \\end{align}

    From the equations above, the crack growths rate can be expressed as a
    function of the stress range. Considering the Paris' law, as in this
    example, gives:

    .. math::

        \\frac{da}{dN} = C \\left( Y(a) \\cdot \\Delta \\sigma \\sqrt{\\pi a}
            \\right)^{m}

    se this crack growth rate together with the cycle step size to
    determine the incremental growth of the crack for this iteration:

    .. math::

        \\Delta a = \\frac{da}{dN} \\cdot \\Delta N

    The new crack size is then calculated by adding the incremental growth
    to the previous crack size:

    .. math::

        a_{j+1} = a_{j} + \\Delta a

    where :math:`a_{j}` is the crack size at the :math:`j`-th cycle.

    The crack is grown iteratively either until the failure condition
    is met or until the stress history is over.

    If failure occurs, the total number of cycles that it took to grow the
    crack to the critical size becomes the predicted life of the part.


    Parameters
    ----------
    stress_range : np.ndarray
        Stress range array.
    count_cycle : np.ndarray
        Number of cycles array.
    slope : np.ndarray
        Slope array.
    intercept : np.ndarray
        Intercept array.
    threshold : float
        Threshold value.
    critical : float
        Critical value.
    crack_geometry : dict
        Crack geometry dictionary. It must contain the following keys:
        - "initial_depth": float
        - "_id": str



    Returns
    -------
    CalcCrackGrowth
        Crack growth object.
    """

    def __init__(
        self,
        stress_range: np.ndarray,
        count_cycle: np.ndarray,
        slope: np.ndarray,
        intercept: np.ndarray,
        threshold: float,
        critical: float,
        crack_type: str,
        crack_geometry: dict,
    ):

        assert intercept.size > 0 and intercept.size == slope.size
        assert np.min(stress_range) >= 0.0
        assert stress_range.size > 0 and stress_range.size == count_cycle.size
        self.stress_range = stress_range
        self.count_cycle = count_cycle
        self.slope = slope
        self.intercept = intercept
        self.threshold = threshold
        self.critical = critical
        self.crack_type = crack_type
        self.crack_geometry = crack_geometry
        cs, sif, factor, failure = self.calc_crack_depth()
        self.crack_depth = cs
        self.sif = sif
        self.geometry_factor = factor
        self.final_cycles = sum(self.count_cycle[: len(self.crack_depth)])
        self.failure = failure

    def __str__(self) -> str:
        """String representation of the object."""
        if self.failure:
            return "Crack growth object: to failure."
        return "Crack growth object: to maximum number of cycles - no failure."

    @property
    def size(self) -> int:
        """Size of the problem."""
        return self.stress_range.size

    @property
    def get_knees_sif(self) -> np.ndarray:
        """Get the SIF values at the knees of the Paris curve.

        If the Paris curve hs more than one (slope, intercept) couple,
        the SIF values at the knees are calculated using the following
        equation:

        .. math::

            \\text{knee}_{i} = \\left(\\frac{\\text{intercept}_{i}}
                {\\text{intercept}_{i+1}}\\right)^{\\frac{1}{
                \\text{slope}_{i+1} - \\text{slope_{i}}}}
        """
        cg_crv_size = self.intercept.size
        knees_sif = np.empty(
            cg_crv_size - 1, dtype=np.float64
        )  # pylint: disable=E1133, C0301
        if cg_crv_size > 1:  # pylint: disable=E1133, C0301
            for i in nb.prange(  # pylint: disable=E1133, C0301
                cg_crv_size - 1
            ):
                knees_sif[i] = (self.intercept[i] / self.intercept[i + 1]) ** (
                    1 / (self.slope[i + 1] - self.slope[i])
                )
        knees_sif = np.hstack(
            (
                np.array([0.9999999999 * self.threshold]),
                knees_sif,
                np.array([self.critical / 0.9999999999]),
            )
        )
        e_msg = (
            "Knee(s) not in between threshold and critical SIF."
            + "\nCheck the definitions of slope, intercept, "
            + "threshold, critical."
        )
        assert np.all(np.diff(knees_sif) > 0), e_msg
        return knees_sif

    def calc_growth_rate(
        self, stress_, crack_depth_
    ) -> Tuple[float, float, float]:
        """Calculate the crack growth rate given the stress and the
        crack size using the Paris' law:

        .. math::

            \\frac{da}{dN} = \\text{intercept} \\, {\\Delta K)^{\\text{slope}}

        Parameters
        ----------
        stress_ : float
            Stress value.
        crack_depth_ : float
            Crack size value.

        Returns
        -------
        Tuple[float, float, float]
            Crack growth rate, stress intensity factor, and geometry factor.
        """

        sif, g_factor = get_sif(
            stress_, crack_depth_, self.crack_type, self.crack_geometry
        )
        knees = self.get_knees_sif
        if sif < self.threshold:
            the_growth_rate = 0.0
        if sif > self.critical:
            the_growth_rate = np.inf
        else:
            idx = np.digitize(np.asarray([sif]), knees, right=True) - 1
            the_growth_rate = (
                self.intercept[idx][0] * sif ** self.slope[idx][0]
            )

        return the_growth_rate, sif, g_factor

    def calc_crack_depth(
        self,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, bool]:
        """Calculate the crack size and the SIF values.

        Using the assigned stress ranges and count-cycles arrays, the final
        crack size is calculated. The method iterates over the stress ranges
        and count-cycles arrays and updates the crack size at each step.

        .. code-block:: python
            :caption: Pseudocode

            crack_depth = [crack_geometry["initial_depth"]]

            sifs = []
            factors = []
            for stress, count_cycle in zip(stress_range, count_cycle):
                for i in range(count_cycle):
                    sifs.append(get_sif(stress,
                                        crack_depth, self.crack_geometry))
                    factors.append(get_geometry_factor(crack_depth,
                                                       crack_type,
                                                       crack_geometry))
                    if sifs[-1] >= critical:
                        return crack_depth, sifs, factors, True
                    if sifs[-1] < threshold:
                        continue
                    crack_depth.append(crack_depth[-1] +
                                       calc_growth_rate(stress, crack_depth))
            return crack_depth, sifs, factors, False

        Returns
        -------
        Tuple[np.ndarray, np.ndarray, np.ndarray, bool]
            Crack size array, SIF array, geometric factor array, failure flag.
        """

        # Initialisation
        the_depth = np.empty(self.size + 1, dtype=np.float64)
        the_depth[0] = self.crack_geometry["initial_depth"]
        the_sif = np.empty(self.size, dtype=np.float64)
        the_factor = np.empty(self.size, dtype=np.float64)
        # Loop
        for i in nb.prange(self.size):  # pylint: disable=E1133
            # the_sif[i], the_factor[i] = get_sif(
            #     self.stress_range[i],
            #     the_depth[i],
            #     self.crack_type,
            #     self.crack_geometry,
            # )
            # if the_sif[i] >= self.critical:
            #     print("Critical SIF reached. Stopping calculation.")
            #     return the_depth[:i], the_sif[:i], the_factor[:i], True
            # if the_sif[i] < self.threshold:
            #     the_depth[i + 1] = the_depth[i]
            #     continue
            incr_depth, the_sif[i], the_factor[i] = self.calc_growth_rate(
                self.stress_range[i], the_depth[i]
            )
            cur_depth = the_depth[i] + incr_depth * self.count_cycle[i]
            if the_factor[i] == 0.0:
                print(
                    "Crack depth is greater than or equal to the thickness."
                    "Stopping calculation."
                )
                return the_depth[:i], the_sif[:i], the_factor[:i], True
            if the_sif[i] >= self.critical:
                print("Critical SIF reached. Stopping calculation.")
                return the_depth[:i], the_sif[:i], the_factor[:i], True
            if the_sif[i] < self.threshold:
                the_depth[i + 1] = the_depth[i]
                continue
            if cur_depth / the_depth[i] > 10:
                w_msg = (
                    "Crack size increased by more than 10x in the"
                    + " last iteration. Stopping calculation."
                )
                print(w_msg)
                return the_depth[:i], the_sif[:i], the_factor[:i], True
            the_depth[i + 1] = cur_depth
        print("Fatigue spectrum applied w/o failure. Stopping calculation")
        return the_depth[:-1], the_sif, the_factor, False


@nb.njit(
    nb.float64(
        nb.float64,
        nb.types.unicode_type,
        nb.types.DictType(nb.types.unicode_type, nb.float64),
    ),
    fastmath=True,
    cache=True,
)
def get_geometry_factor(  # pragma: no cover
    crack_depth: float,
    crack_type: str,
    crack_geometry: dict,  # pylint: disable=W0613, C0301
) -> float:
    """Get the geometric factor. This should be a function of the crack
    size. The default value is 1.0."""
    # Placeholder for future implementation
    if crack_type == "HOL_CYL_01":
        return f_hol_cyl_01(crack_depth, crack_geometry)
    if crack_type == "INF_SUR_00":
        return 1.0

    e_msg = "Geometric factor not implemented for assigned crack type."
    raise NotImplementedError(e_msg)


@nb.njit(
    nb.types.UniTuple(nb.float64, 2)(
        nb.float64,
        nb.float64,
        nb.types.unicode_type,
        nb.types.DictType(nb.types.unicode_type, nb.float64),
    ),
    fastmath=True,
    cache=True,
)
def get_sif(  # pragma: no cover
    stress_: float,
    crack_depth_: float,
    crack_type: str,
    crack_geometry: Dict[list, float],
) -> Tuple[float, float]:
    """Get the SIF value given the stress and the crack size. The
    method calculates the SIF value according to the following
    equation:

    .. math::

        SIF = Y(a) \\cdot \\Delta \\sigma \\cdot \\sqrt{\\pi a}

    Parameters
    ----------
    stress_ : float
        Stress value.
    crack_depth_ : float
        Crack size value.
    crack_type : str
        Crack type.
    crack_geometry : dict
        Crack geometry dictionary.

    Returns
    -------
    Tuple[float, float]
        SIF value, geometric factor.
    """
    y_1 = get_geometry_factor(crack_depth_, crack_type, crack_geometry)
    return stress_ * np.sqrt(np.pi * crack_depth_) * y_1, y_1


def get_crack_growth(  # pragma: no cover
    cycle_count: CycleCount,
    cg_curve: ParisCurve,
    crack_geometry: Type[ACG],
    express_mode: bool = False,
) -> CalcCrackGrowth:
    """Calculate the crack size by numerically integrating the crack growth
    curve. In this example we will consider the Paris' law, but the function
    can be used with any other crack growth curve.

    .. math::

        \\begin{align}
        \\begin{cases}
            \\frac{da}{dN} & = C \\left( \\Delta K \\right)^{m} \\\\
            \\Delta K & = Y(a) \\cdot \\Delta \\sigma \\sqrt{\\pi a} \\\\
        \\end{cases}
        \\end{align}

    From the equations above, the crack growths rate can be expressed as a
    function of the stress range. Considering the Paris' law, as in this
    example, gives:

    .. math::

        \\frac{da}{dN} = C \\left( Y(a) \\cdot \\Delta \\sigma \\sqrt{\\pi a}
            \\right)^{m}

    se this crack growth rate together with the cycle step size to determine
    the incremental growth of the crack for this iteration:

    .. math::

        \\Delta a = \\frac{da}{dN} \\cdot \\Delta N

    The new crack size is then calculated by adding the incremental growth to
    the previous crack size:

    .. math::

        a_{j+1} = a_{j} + \\Delta a

    where :math:`a_{j}` is the crack size at the :math:`j`-th cycle.

    The crack is grown iteratively either until the failure condition is met
    or until the stress history is over.

    If failure occurs, the total number of cycles that it took to grow the
    crack to the critical size becomes the predicted life of the part.

    Parameters
    ----------
    cycle_count : py_fatigue.CycleCount
        Cycle count object.
    cg_curve : py_fatigue.ParisCurve
        The crack growth curve
    crack_geometry : Type[ACG]
        The initial crack geometry
    express_mode : bool, optional
        If True, the crack growth is calculated using the express mode. This
        mode is faster but 'slightly' less accurate. The default is False.

    Returns
    -------
    CalcCrackGrowth
        The crack growth calculation object
    """
    # Handle the stress ranges and cycle counts as they might be clustered
    # in a matrix and we need to iterate over single or "little" cycles
    # within the same stress range block
    if cycle_count.unit not in cg_curve.unit:
        e_msg = (
            f"Cycle count unit ({cycle_count.unit}) and crack growth "
            f"curve unit ({cg_curve.unit}) are not compatible."
        )
        raise ValueError(e_msg)
    if np.max(cycle_count.count_cycle) > 1:
        if express_mode:
            splts = [
                10 ** int(np.log10(cycle)) for cycle in cycle_count.count_cycle
            ]
        else:
            splts = [int(cycle) for cycle in cycle_count.count_cycle]
        splt_cyc = [
            split(int(cycle), n) if int(cycle) > 0 else [cycle]
            for cycle, n in zip(cycle_count.count_cycle, splts)
        ]
        splt_len = [len(_) for _ in splt_cyc]

        count_cycle = np.asarray(
            [float(item) for sublist in splt_cyc for item in sublist]
        )
        stress_range = np.repeat(cycle_count.stress_range, splt_len)
    else:
        stress_range = cycle_count.stress_range
        count_cycle = cycle_count.count_cycle

    # Handle the geometry

    crack_type = str(crack_geometry._id)
    if crack_type not in ["HOL_CYL_01", "INF_SUR_00"]:
        raise ValueError("Unsupported crack geometry")
    geometry = to_numba_dict(crack_geometry.__dict__)  # type: ignore
    cg = CalcCrackGrowth(
        stress_range,
        count_cycle,
        cg_curve.slope,
        cg_curve.intercept,
        float(cg_curve.threshold),
        float(cg_curve.critical),
        crack_type,
        geometry,
    )
    # crack_geometry.geometry_factor = cg.geometry_factor
    # crack_geometry.crack_depth_ = cg.crack_depth
    # crack_geometry.sif_ = cg.sif
    # crack_geometry.count_cycle_ = np.cumsum(
    #     cg.count_cycle[: len(cg.crack_depth)]
    # )
    # crack_geometry.final_cycles_ = cg.final_cycles
    return cg  # crack_geometry


@pd.api.extensions.register_dataframe_accessor("cg")
class CrackGrowth:
    """Accessor for the crack growth calculation

    Parameters
    ----------
    pandas_obj : pd.DataFrame
        The pandas dataframe object that will be extended
        with the crack growth calculation results. The original dataframe must
        have the following columns:
        - stress_range
        - count_cycle
        - mean_stress
        The dataframe will be extended with the following columns:
        - crack_depth
        - sif
        - cumul_cycles
        - geometry_factor
    """

    def __init__(self, pandas_obj):
        # self._validate(pandas_obj)
        self._obj = pandas_obj
        self.cg_curve = None
        self.crack_geometry = None
        self.final_cycles = None

    @staticmethod
    def _validate(obj):
        """Validate the input DataFrame. Raise an error if the input
        DataFrame does not contain the right columns.
        """
        if {
            "crack_depth",
        }.issubset(obj.columns):
            e_msg = "'crack_depth' already calculated"
            raise AttributeError(e_msg)

        if not {"count_cycle", "mean_stress", "stress_range"}.issubset(
            obj.columns
        ):
            e_msg = (
                "Must have 'count_cycle', 'mean_stress' and 'stress_range'."
            )
            raise AttributeError(e_msg)

    def calc_growth(
        self,
        cg_curve: ParisCurve,
        crack_geometry: Type[ACG],
        express_mode: bool = False,
    ):
        """Calculate the crack_propagation of the given crack geometry.

        Parameters
        ----------
        cg_curve : py_fatigue.ParisCurve
            The crack growth curve
        crack_geometry : Type[ACG]
            The initial crack geometry
        express_mode : bool, optional
            If True, the crack propagation runs 'express', by default False

        Returns
        -------
        DataFrame
            The DataFrame with the results of the crack_growth analysis.
        """

        self._validate(self._obj)

        # Handle the stress ranges and cycle counts as they might be clustered
        # in a matrix and we need to iterate over single or "little" cycles
        # within the same stress range block

        if np.max(self._obj["count_cycle"]) > 1:
            if express_mode:
                splts = [
                    10 ** int(np.log10(cycle))
                    for cycle in self._obj["count_cycle"].values
                ]
            else:
                splts = [
                    int(cycle) for cycle in self._obj["count_cycle"].values
                ]
            splt_cyc = [
                split(int(cycle), n) if int(cycle) > 0 else [cycle]
                for cycle, n in zip(self._obj["count_cycle"].values, splts)
            ]
            splt_len = [len(_) for _ in splt_cyc]

            count_cycle = np.asarray(
                [float(item) for sublist in splt_cyc for item in sublist]
            )
            stress_range = np.repeat(
                self._obj["stress_range"].values, splt_len
            )
        else:
            stress_range = self._obj["stress_range"].values
            count_cycle = self._obj["count_cycle"].values

        # Handle the geometry
        crack_type = str(crack_geometry._id)
        if crack_type not in ["HOL_CYL_01", "INF_SUR_00"]:
            raise ValueError("Unsupported crack geometry")
        geometry = to_numba_dict(crack_geometry.__dict__)  # type: ignore

        cg_ = CalcCrackGrowth(
            stress_range,
            count_cycle,
            cg_curve.slope,
            cg_curve.intercept,
            float(cg_curve.threshold),
            float(cg_curve.critical),
            crack_type,
            geometry,
        )

        self._obj.final_cycles = cg_.final_cycles
        self.final_cycles = cg_.final_cycles

        cumul_cycle = np.cumsum(cg_.count_cycle[: len(cg_.crack_depth)])
        # manage the fact that clustered cycles are split in cg analysys
        # while the df keeps them clustered
        d_l = len(self._obj) - len(cg_.crack_depth)
        if d_l >= 0:
            self._obj["crack_depth"] = np.concatenate(
                (cg_.crack_depth, np.full(d_l, np.nan))
            )
            self._obj["sif"] = np.concatenate((cg_.sif, np.full(d_l, np.nan)))
            self._obj["cumul_cycle"] = np.concatenate(
                (cumul_cycle, np.full(d_l, np.nan))
            )
            self._obj["geometry_factor"] = np.concatenate(
                (cg_.geometry_factor, np.full(d_l, np.nan))
            )
        else:
            idx = np.asarray(self._obj["count_cycle"].cumsum(), dtype=int) - 1
            self._obj["crack_depth"] = cg_.crack_depth[idx]
            self._obj["sif"] = cg_.sif[idx]
            self._obj["cumul_cycle"] = cumul_cycle[idx]
            self._obj["geometry_factor"] = cg_.geometry_factor[idx]
        self._obj.cg_curve = cg_curve
        self.cg_curve = cg_curve
        self._obj.crack_geometry = crack_geometry
        self.crack_geometry = crack_geometry
        return self._obj



"""
The :mod:`py_fatigue.utils` module collects all the utility functions
and classes.
"""

# Packages from the Python Standard Library
from __future__ import annotations
from dataclasses import dataclass
from typing import (
    Any,
    Generator,
    List,
    Tuple,
    cast,
    Callable,
    ClassVar,
    Generic,
    NamedTuple,
    Optional,
    TypeVar,
    Union,
)
import copy

# Packages from non-standard libraries
# from pydantic.fields import ModelField
import matplotlib
import matplotlib.pyplot as plt
import numba as nb
import numpy as np


def inplacify(method: Callable) -> Callable:  # pragma: no cover
    """
    Make a method inplace.

    Parameters
    ----------
    method : Callable
        The method to make inplace.

    Returns
    -------
    Callable
        The inplace method.
    """

    def wrap(self, *args, **kwargs):
        inplace = kwargs.pop("inplace", False)
        if inplace:
            return method(self, *args, **kwargs)
        return method(copy.copy(self), *args, **kwargs)

    return wrap


def make_axes(
    fig: Optional[matplotlib.figure.Figure] = None,  # type: ignore
    ax: Optional[matplotlib.axes.Axes] = None,  # type: ignore
) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:  # type: ignore  # pragma: no cover  # pylint: disable=C0301  # noqa: E501
    """Check if a figure and axes are provided, and if not, create them.

    Parameters
    ----------
    fig : Optional[matplotlib.figure.Figure], optional
        The figure instance, by default None
    ax : Optional[matplotlib.axes.Axes], optional
        The axes instance, by default None

    Returns
    -------
    tuple
        The figure and axes instances

    Raises
    ------
    TypeError
        If fig is not a matplotlib.figure.Figure instance
    """
    if fig is None:
        fig, axes = plt.subplots()
    else:
        if not isinstance(fig, matplotlib.figure.Figure):  # type: ignore
            raise TypeError("fig must be a matplotlib.figure.Figure instance")
        if ax is None:
            axes = fig.gca()
        else:
            axes = ax
    return fig, axes


class IntInt(NamedTuple):
    """Represents a pair of integers, namely the number of bins for
    mean and ranges

    Parameters
    ----------
    NamedTuple : NamedTuple
    """

    mean_bin_nr: int
    range_bin_nr: int


class ArrayArray(NamedTuple):  # pragma: no cover
    """Represents a pair of arrays, namely the mean and range bin edge
    values

    Parameters
    ----------
    NamedTuple : NamedTuple
    """

    mean_bin_edges: np.ndarray
    range_bin_edges: np.ndarray


class IntArray(NamedTuple):
    """Represents an integer and an array, namely the number of bins
    for mean and range bin edge values

    Parameters
    ----------
    NamedTuple : NamedTuple
    """

    mean_bin_nr: int
    range_bin_edges: np.ndarray


class ArrayInt(NamedTuple):
    """Represents an array and an integer, namely the number of bins
    for range and mean bin edge values

    Parameters
    ----------
    NamedTuple : NamedTuple
    """

    mean_bin_edges: np.ndarray
    range_bin_edges: int


UnionOfIntArray = Union[
    int, np.ndarray, IntInt, ArrayArray, IntArray, ArrayInt
]


@dataclass(repr=False)
class FatigueStress:
    """Fatigue stress class. Mean stress, range, maximum, and minimum
    can inherit from this class that defines the following properties:

        - `bin_edges`: the bin edges for the histogram
        - `bin_centers`: the bin centers for the histogram
        - `full`: the full cycles
        - `full_counts`: the counts of the full cycles
        - `half`: the half cycles
        - `half_counts`: the counts of the half cycles
        - `bins_idx`: the index of the full cycles bins
        - `binned_values`: the binned full cycles
    """

    category: ClassVar[str]
    _counts: np.ndarray
    _values: np.ndarray
    bin_width: float
    _bin_lb: Optional[float] = None
    _bin_ub: Optional[float] = None
    # @properties
    # bin_edges: np.ndarray
    # bin_centers: np.ndarray
    # full: np.ndarray
    # half: np.ndarray
    # half_counts: np.ndarray
    # bins_idx: np.ndarray
    # binned_values: np.ndarray

    def __post_init__(self):
        if len(self._counts) == 0 or len(self._values) == 0:
            raise ValueError("No data provided")
        if len(self._counts) != len(self._values):
            raise ValueError("counts and values must have the same length")
        self._counts = np.asarray(self._counts)
        self._values = np.asarray(self._values)
        if self._bin_lb is None:
            self._bin_lb = (
                round(min(self._values) / self.bin_width) * self.bin_width
                - self.bin_width / 2
            )
        if self._bin_ub is None:
            self._bin_ub = bin_upper_bound(
                self._bin_lb,  # type: ignore
                self.bin_width,
                max(self.full) if len(self.full) > 0 else max(self._values),
            )

    @property
    def bin_lb(self) -> float:
        """The lower bound of the bins. Casting resolves mypy error"""
        return cast(float, self._bin_lb)

    @bin_lb.setter
    def bin_lb(self, value: float) -> None:
        self._bin_lb = value

    @property
    def bin_ub(self) -> float:
        """The upper bound of the bins. Casting resolves mypy error"""
        return cast(float, self._bin_ub)

    @bin_ub.setter
    def bin_ub(self, value: float) -> None:
        self._bin_ub = value

    @property
    def counts(self) -> np.ndarray:
        """The number of cycles.

        Returns
        -------
        counts : np.ndarray
        """
        return self._counts

    # val setter function
    @counts.setter
    def counts(self, val: np.ndarray) -> None:
        """Setter for the counts.

        Parameters
        ----------
        val : np.ndarray
            The new counts.

        Returns
        -------
        counts : np.ndarray
        """
        self._counts = val

    @property
    def values(self) -> np.ndarray:
        """The stress values.

        Returns
        -------
        values : np.ndarray
        """
        return self._values

    # val setter function
    @values.setter
    def values(self, val: np.ndarray) -> None:
        """Setter for the values.

        Parameters
        ----------
        val : np.ndarray
            The new values.
        """
        self._values = val

    @property
    def bin_edges(self) -> np.ndarray:
        """The bin edges.

        Returns
        -------
        bin_edges : np.ndarray
        """
        return calc_bin_edges(self.bin_lb, self.bin_ub, self.bin_width)

    @property
    def bin_centers(self):
        """The bin centers.

        Returns
        -------
        bin_centers : np.ndarray
        """
        return self.bin_edges[:-1] + self.bin_width / 2

    @property
    def full(self):
        """The full cycles stresses.

        Returns
        -------
        full : np.ndarray
        """
        return calc_full_cycles(self.values, self.counts)

    @property
    def half(self):
        """The half cycles stresses.

        Returns
        -------
        half : np.ndarray
        """
        return calc_half_cycles(self.values, self.counts)

    @property
    def bins_idx(self):
        """The index of the full cycles bins.

        Returns
        -------
        bins_idx : np.ndarray
        """
        digitized = np.digitize(self.full, self.bin_edges, right=True)
        digitized[np.where(digitized > 0)] -= 1
        return digitized

    @property
    def binned_values(self):
        """The binned full cycles.

        Returns
        -------
        binned_values : np.ndarray
        """
        return self.bin_centers[self.bins_idx]


def calc_bin_edges(
    bin_lb: float, bin_ub: float, bin_width: float
) -> np.ndarray:
    """Calculate the bin edges for a fatigue histogram using the given
    bin width, lower bound and upper bound.

    Parameters
    ----------
    bin_lb : float
        Lowest bin edge value
    bin_ub : float
        Uppest bin edge value
    bin_width : float
        Bin width

    Returns
    -------
    np.ndarray
        Bin edges
    """
    bin_ub = bin_upper_bound(bin_lb, bin_width, bin_ub)

    n_bins = int(np.round((bin_ub + bin_width - bin_lb) / bin_width))

    bin_edges = np.around(
        np.linspace(start=bin_lb, stop=bin_ub, num=n_bins),
        decimals=5,
    )
    return bin_edges


def bin_upper_bound(
    bin_lower_bound: float, bin_width: float, max_val: float
) -> float:
    """Returns the upper bound of the bin edges, given the lower bound,
    bin width and maximum value of the array. If the maximum value is
    not a multiple of the bin width, the upper bound is rounded up to
    the next multiple of the bin width.
    Also, if the maximum value is smaller than the lower bound, the
    upper bound is set to the lower bound.

    Parameters
    ----------
    bin_lower_bound : float
        Left edge of the first bin
    bin_width : float
        Bin width
    max_val : float
        Maximum value of the array to be binned

    Returns
    -------
    float
        The upper bound of the bin edges
    """
    return bin_lower_bound + bin_width * np.ceil(
        (max_val - bin_lower_bound) / bin_width
    )


def split_full_cycles_and_residuals(count_cycle: np.ndarray) -> tuple:
    """
    Returns the :term:`stress ranges<Stress Ranges>` or
    :term:`mean stresses<Mean stress>` from only the full cycles.
    """

    splt = np.asarray(np.modf(count_cycle)).T
    full_idx = np.where((splt[:, 1] % 1 == 0) & (splt[:, 1] > 0))[0]
    half_idx = np.where((splt[:, 0] == 0.5) & (splt[:, 0] > 0))[0]
    # zero_idx = np.where((splt[:, 0] == 0) & (splt[:, 1] == 0))[0]
    # full_idx = np.sort(np.hstack([full_idx, zero_idx]))
    # half_idx = np.sort(np.hstack([half_idx, zero_idx]))
    return full_idx, half_idx


def calc_full_cycles(
    stress: np.ndarray, count_cycle: np.ndarray, debug_mode: bool = False
) -> np.ndarray:
    """
    Returns the :term:`stress ranges<Stress Ranges>` or
    :term:`mean stresses<Mean stress>` from only the full cycles.
    """
    full_idx, _ = split_full_cycles_and_residuals(count_cycle)
    # integer_count = np.delete(count_cycle, half_idx).astype(int)
    integer_count = count_cycle[full_idx].astype(int)
    integer_stress = stress[full_idx]
    if debug_mode:
        print("full cycles idx", full_idx)
        print("count_cycle[full_idx]", integer_count)
        print("stress[full_idx]", integer_stress)
        print("len integer_count", len(integer_count))
        print("len integer_stress", len(integer_stress))
    return np.repeat(integer_stress, integer_count)


def calc_half_cycles(
    stress: np.ndarray, count_cycle: np.ndarray, debug_mode: bool = False
) -> np.ndarray:
    """
    Returns the :term:`stress ranges<Stress Ranges>` or
    :term:`mean stresses<Mean stress>` from only the half cycles.
    """
    _, half_idx = split_full_cycles_and_residuals(count_cycle)

    half_stress = stress[half_idx]
    if debug_mode:
        half_count = count_cycle[half_idx]
        print("half cycles idx", half_idx)
        print("count_cycle[half_idx]", half_count)
        print("len half_count", len(half_count))
        print("len half_stress", len(half_stress))
    return half_stress


def compare(a: Any, b: Any) -> bool:
    """Compare two objects.

    Parameters
    ----------
    a : Any
    b : Any

    Returns
    -------
    bool
        True if the a and b are equal, False otherwise.
    """
    # print(f"a: {a}, b: {b}")
    if not hasattr(a, "__iter__") or isinstance(a, str):
        return a == b

    try:
        if not len(a) == len(b):
            return False
        if isinstance(a, np.ndarray):
            return np.array_equal(a, b)
        if isinstance(a, dict):
            return all(compare(v, b[k]) for k, v in a.items())
        return all(compare(aa, bb) for aa, bb in zip(a, b))
    except (TypeError, KeyError):
        return False


def split(what: int, by: int):
    """Split an integer into a list of integers.

    >>> split(5, 2)
    [2, 3]
    >>> split(20, 7)
    [3, 3, 3, 3, 3, 3, 2]

    Parameters
    ----------
    what : int
        The integer to split
    by : int
        The number of integers to split the integer into

    Returns
    -------
    list
        The list of integers
    """
    return [what // by + 1] * (what % by) + [what // by] * (by - what % by)


JSON_ENCODERS = {np.ndarray: lambda arr: arr.tolist()}

DType = TypeVar("DType")


class TypedArray(np.ndarray, Generic[DType]):
    """Wrapper class for numpy arrays that stores and validates type
    information. This can be used in place of a numpy array, but when
    used in a pydantic BaseModel or with pydantic.validate_arguments,
    its dtype will be *coerced* at runtime to the declared type.
    """

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val, field):
        """Validate the value of the field."""
        dtype_field = field.sub_fields[0]  # type: ignore
        actual_dtype = dtype_field.type_.__args__[0]
        # If numpy cannot create an array with the request dtype,
        # an error will be raised and correctly bubbled up.
        np_array = np.array(val, dtype=actual_dtype)
        return np_array


def to_numba_dict(
    data: dict, key_type: type = str, val_type: type = float
) -> nb.types.DictType:
    """Converts a dictionary to a numba typed dict, provided the output
    key and value types.

    Parameters
    ----------
    data : dict
        The dictionary to be converted.
    key_type : type, optional
        The key type, by default str
    val_type : type, optional
        The type of the values, by default float

    Returns
    -------
    nb.types.DictType
        The numba typed dictionary.
    """
    dct = nb.typed.Dict.empty(
        key_type=nb.types.string,
        value_type=nb.float64,
    )
    for key, value in data.items():
        if not isinstance(value, val_type) or not isinstance(key, key_type):
            continue
        dct[key] = value
    return dct


def chunks(lst: list[Any], n: int) -> Generator[list[Any], None, None]:
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def calc_slope_intercept(
    x: Union[np.ndarray, List[float]],
    y: Union[np.ndarray, List[float]],
) -> tuple[np.ndarray, np.ndarray]:
    """
    Given a series of (x,y) knee point coordinates, return the slope and
    intercept of the line that passes through the knee point and the previous
    point.

    Equation of the line:

    .. math::

        y = mx + c

    where:

    - | :math:`m` is the slope of the line, calculated as
      | :math:`\\frac{y_2 - y_1}{x_2 - x_1}`
    - | :math:`c` is the intercept of the line, calculated as
      | :math:`\\exp(y_2 - m \\times x_2)`

    Parameters
    ----------
    x : ArrayLike
        The x-coordinates of the knee points.
    y : ArrayLike
        The y-coordinates of the knee points.
    Returns
    -------
    tuple[ArrayLike, ArrayLike]
        The slope and intercept of the line that passes through the knee
        point and the previous point.
    """
    assert len(x) == len(y), "x and y must have the same length"
    assert len(x) >= 2, "x and y must have at least two points"
    x = np.sort(np.array(x))
    y = np.array(y)[np.argsort(x)]

    # Check that the y values are strictly increasing
    # assert np.all(np.diff(y) > 0), "x values must be strictly increasing"
    # Find the slope and intercept for each pair of points, meaning that
    # x and y have at least to be of length 2.
    slopes = (y[1:] - y[:-1]) / (x[1:] - x[:-1])  # type: ignore
    intercepts = y[1:] - slopes * x[1:]
    return slopes, intercepts



# -*- coding: utf-8 -*-

r"""The following code is based on the histogram function from the numpy
package `[4]`_.

.. _[4]: https://numpy.org/doc/stable/reference/generated/numpy.histogram2d
"""

# standard imports
from __future__ import annotations
import itertools
from typing import Callable, List, Optional, Sequence, Union, TypeVar

# non-standard imports
import matplotlib
import numpy as np
import numpy.typing as npt

from .rainflow import rainflow as calc_rainflow
from ..utils import make_axes
from ..mean_stress import MeanStress
from ..stress_range import StressRange

__all__ = [
    "make_axes",
    "make_histogram",
    "rainflow_binner",
    "binned_rainflow",
    "xy_hist2d",
]


FloatArray = TypeVar("FloatArray", bound=npt.NDArray[np.floating])


def make_histogram(
    mean_stress: np.ndarray,
    stress_range: np.ndarray,
    bins: Union[int, Sequence[int]] = 10,
    bin_range: Optional[np.ndarray] = None,
    normed: bool = False,
    weights: Optional[np.ndarray] = None,
) -> tuple:
    """
    Make a scattered-histogram plot.

    Parameters
    ----------
    mean_stress, stress_range : array_like, shape (n, )
        Input data
    bins : int or array_like or [int, int] or [array, array], optional
        The bin specification:
          * If int, the number of bins for the two dimensions (nx=ny=bins).
          * If array_like, the bin edges for the two dimensions
            (stress_range_edges=mean_stress_edges=bins).
          * If [int, int], the number of bins in each dimension
            (n_stress_range, n_mean_stress = bins).
          * If [array, array], the bin edges in each dimension
            (stress_range_edges, mean_stress_edges = bins).
          * A combination [int, array] or [array, int], where int
            is the number of bins and array is the bin edges.
    bin_range : array_like, shape(2,2), optional
        The leftmost and rightmost edges of the bins along each dimension
        (if not specified explicitly in the `bins` parameters):
        ``[[stress_range_min, stress_range_max],
        [mean_stress_min, mean_stress_max]]``. All values outside of this
        range will be considered outliers and not tallied in the histogram.
    normed : bool, optional
        If False, returns the number of samples in each bin. If True,
        returns the bin density ``bin_count / sample_count / bin_area``.
    weights : array_like, shape(N,), optional
        An array of values ``w_i`` weighing each sample
        ``(stress_range_i, mean_stress_i)``. Weights are normalized to 1
        if `normed` is True. If `normed` is False, the values of the returned
        histogram are equal to the sum of the weights belonging to the samples
        falling into each bin.

    Returns
    -------
    hist : `np.ndarray`, shape(m,m)
        Histogram matrix.
    hist_edges : ArrayArray, (shape(m+1,), shape(m+1,))
        Histogram bin edges.
    """

    if len(mean_stress) != len(stress_range):
        raise ValueError("Input arrays must be the same length")

    try:
        hist, mean_e, range_e = np.histogram2d(
            mean_stress,
            stress_range,
            bins=bins,
            range=bin_range,
            normed=normed,
            weights=weights,
        )
    except TypeError as type_error:
        if "normed" in str(type_error):
            hist, mean_e, range_e = np.histogram2d(
                mean_stress,
                stress_range,
                bins=bins,
                range=bin_range,
                weights=weights,
                density=normed,
            )
    hist_bin_edges = (mean_e, range_e)
    return hist, hist_bin_edges


def map_hist(
    x: np.ndarray,
    y: np.ndarray,
    h: np.ndarray,
    hist_bin_edges: tuple,
) -> np.ndarray:
    """Maps the histogram to the original data.

    Parameters
    ----------
    x : np.ndarray
        x-axis of the original data
    y : np.ndarray
        y-axis of the original data
    h : np.ndarray
        histogram
    hist_bin_edges : tuple
        histogram bin edges

    Returns
    -------
    np.ndarray
        mapped histogram
    """
    x_i = np.digitize(x, hist_bin_edges[0])
    x_i[np.where(x_i > 0)] -= 1
    y_i = np.digitize(y, hist_bin_edges[1])
    y_i[np.where(y_i > 0)] -= 1
    inds = np.ravel_multi_index(
        (x_i, y_i),
        (len(hist_bin_edges[0]) - 1, len(hist_bin_edges[1]) - 1),
        mode="clip",
    )
    vals = h.flatten()[inds]
    bads = (
        (x < hist_bin_edges[0][0])
        | (x > hist_bin_edges[0][-1])
        | (y < hist_bin_edges[1][0])
        | (y > hist_bin_edges[1][-1])
    )
    vals[bads] = np.nan
    return vals


def xy_hist2d(
    x: FloatArray,
    y: FloatArray,
    mode: str = "mountain",
    bins: Union[int, Sequence[int]] = 10,
    bin_range: Optional[FloatArray] = None,
    normed: bool = False,
    weights: Optional[FloatArray] = None,  # np.histogram2d args
    fig: Optional[matplotlib.figure.Figure] = None,
    ax: Optional[matplotlib.collections.PathCollection] = None,
    dens_func: Optional[Callable] = None,
    **kwargs: dict,
) -> tuple:
    """
    Make a scattered-histogram plot.

    Parameters
    ----------
    x, y : array_like, shape (n, )
        Input data
    marker_size : int
        Size of the markers
    mode: [None | 'mountain' | 'valley' | 'clip']
        Possible values are:
            - None : The points are plotted as one scatter object, in the
                order in-which they are specified at input.
            - 'mountain' : The points are sorted/plotted in the order of
                the number of points in their 'bin'. This means that points
                in the highest density will be plotted on-top of others. This
                cleans-up the edges a bit, the points near the edges will
                overlap.
            - 'valley' : The reverse order of 'mountain'. The low density
                bins are plotted on top of the high ones.

    bins : int or array_like or [int, int] or [array, array], optional
        The bin specification:
            - If int, the number of bins for the two dimensions (nx=ny=bins).
            - If array_like, the bin edges for the two dimensions
                (x_edges = y_edges = bins).
            - If [int, int], the number of bins in each dimension
                (n_x, n_y = bins).
            - If [array, array], the bin edges in each dimension
                (x_edges, y_edges = bins).
            - A combination [int, array] or [array, int], where int
                is the number of bins and array is the bin edges.

    bin_range : array_like, shape(2,2), optional
        The leftmost and rightmost edges of the bins along each dimension
        (if not specified explicitly in the `bins` parameters):
        ``[[x_min, x_max], [y_min, y_max]]``. \
        All values outside of this range will be considered outliers
        and not tallied in the histogram.
    normed : bool, optional
        If False, returns the number of samples in each bin. If True,
        returns the bin density ``bin_count / sample_count / bin_area``.
    weights : array_like, shape(N,), optional
        An array of values ``w_i`` weighing each sample ``(y_i, x_i)``.
        Weights are normalized to 1 if `normed` is True. If `normed` is
        False, the values of the returned histogram are equal to the sum of
        the weights belonging to the samples falling into each bin.
    fig : a figure instance to add.
    ax : an axes instance to plot into.
    dens_func : function or callable (default: None)
        A function that modifies (inputs and returns) the dens
        values (e.g., np.log10). The default is to not modify the
        values.
    x_label : str, optional
        Label for the x-axis
    y_label : str, optional
        Label for the y-axis
    cbar_label : str, optional
        Label for the colorbar
    kwargs : these are all passed on to scatter.

    Returns
    -------
    figure, paths : (
        `~matplotlib.figure.Figure`,
        `~matplotlib.collections.PathCollection`
    )
        The figure with scatter instance.
    """
    fig, axes = make_axes(fig, ax)
    hist, hist_bin_edges = make_histogram(
        x, y, bins=bins, bin_range=bin_range, normed=normed, weights=weights
    )
    dens = map_hist(x, y, hist, hist_bin_edges)
    if dens_func is not None:
        dens = dens_func(dens)
    # iorder = np.empty(0, dtype=float)
    if mode == "mountain":
        iorder = np.argsort(dens)
    elif mode == "valley":
        iorder = np.argsort(dens)[::-1]
    else:
        raise ValueError("mode must be one of [None, 'mountain', 'valley']")
    x = x[iorder]
    y = y[iorder]
    dens = dens[iorder]
    im = axes.scatter(
        x,
        y,
        # edgecolors="#CCC",
        # linewidths=0.5,
        c=dens,
        **kwargs,
    )
    return fig, axes, im


def rainflow_binner(
    ranges: StressRange,
    _means: Optional[MeanStress],
    damage_tolerance_for_binning: float = 1e-3,
    damage_exponent: float = 5.0,
    max_consecutive_zeros: int = 12,
    round_decimals: int = 4,
    debug_mode: bool = True,
) -> dict:
    """Binning the rainflow cycles into mean-range bins up to a certain
    stress range which is based on the damage tolerance index. All the
    cycles beyond such stress range are saved separately for accuracy
    reasons as "large cycles".

    By default, large cycles are calculated from all the cycles
    accounting for more than 0.1% of the total damage and that have a
    discrepancy of more than 0.1% with respect to the damage that would
    be obtained from the bin center.

    Parameters
    ----------
    ranges : StressRange
        The stress range object.
    _means : Optional[MeanStress], optional
        The mean stress object. If not provided, the mean stress is
        set to zero.
    damage_tolerance_for_binning : float, optional
        tolerance for the damage when binning large cycles
        separately, by default 1e-3
    damage_exponent : float, optional
        exponent for the damage when binning large cycles separately,
        by default 5.0
    max_consecutive_zeros : int, optional
        maximum number of consecutive zeros before saving a cycle to
        large cycles array. This quantity is introduced to save storage
        space, by default 8
    round_decimals : int, optional
        number of decimals to round the output, by default 4
    debug_mode : bool, optional
        if True, print debug messages, by default False

    Returns
    -------
    out_dct : dict
        A dictionary with the following keys:
        - "nr_small_cycles": number of small cycles
        - "range_bin_lower_bound": lower bound of the range bin
        - "range_bin_width": width of the range bins
        - "mean_bin_lower_bound": lower bound of the mean bin
        - "mean_bin_width": width of the mean bins
        - "hist": histogram of the rainflow cycles
        - "lg_c": large cycles
        - "res": residuals
        - "res_sig": residuals sequence

    See Also
    --------
    make_histogram
    """
    # 1) Type of analysis: if with mean or without mean.
    # If no mean stress is desired, the method creates a dummy array of
    # zero mean stresses.
    no_mean_stress = True
    means = MeanStress(
        ranges.counts,
        np.zeros(len(ranges.values)),
        1,
    )
    if _means is not None:
        means = _means
        no_mean_stress = False

    # 2) Removing mean-range pairs below threshold (lowest bound)
    ranges, means, nr_small_cycles = rainflow_threshold(ranges, means)
    if debug_mode:  # pragma: no cover
        # a) Full cycles
        print("\033[1mInitial values\033[0m\n--------------")
        print("count cycle", ranges.counts, "\n")
        print("mean stress", means.values, "\n")
        print("stress range", ranges.values, "\n")
        # b) Full cycles
        print("\033[1mFull cycles\033[0m\n-----------")
        # print("count full",means.counts)
        print("mean full", means.full, "\n")
        print("range full", ranges.full, "\n")
        print("len mean full", len(means.full))
        print("len range full", len(ranges.full), "\n")
        # c) Residuals
        print("\033[1mResiduals\033[0m\n---------")
        print("Mean :", means.half, "\n")
        print("Range:", ranges.half, "\n")
        print("\033[1mLoop over rainflow-counted data\033[0m")
        print("-------------------------------")
    # 3) Histogram bin edges (all these attributes are in FatigueStress)
    # 3.a) Mean - lower bound
    # 3.b) Mean - upper bound
    # 3.c) Mean - edges
    # 3.d) Range - upper bound
    # 3.e) Range - edges

    # 4) Histogram
    # _, hist_bin_edges = make_histogram(
    #     means.full,
    #     ranges.full,
    #     bins=(means.bin_edges, ranges.bin_edges)
    # )

    # # Test on the edges output from numpy.histogram
    # np.testing.assert_array_equal(mean_bin_edges, hist_bin_edges[0])
    # np.testing.assert_array_equal(range_bin_edges, hist_bin_edges[1])

    # 5) Indices of the bins for mean and range using numpy.digitize, ref.
    # https://numpy.org/doc/stable/reference/generated/numpy.digitize.html
    # -------------------------------------------------------------------------

    binned_sorted = sorted(
        zip(
            means.bins_idx,
            ranges.bins_idx,
            means.binned_values,
            ranges.binned_values,
            means.full,
            ranges.full,
        ),
        reverse=False,
    )

    # 6) Initialising the output dictionary
    out_dct = initialise_binned_rainflow_dict(
        ranges, means, nr_small_cycles, no_mean_stress
    )
    if len(ranges.full) == 0:
        return _build_binned_dict(
            ranges, means, [], [], round_decimals, no_mean_stress, out_dct
        )
    # 6) Reference damage for large cycles
    max_damage = max(ranges.full) ** damage_exponent
    # 7) Initialising output variabies: large cycles and histogram
    lg_c = []
    # the_hist: List[List[int]] = [[] for _ in range(len(means.bin_centers))]
    _hist = np.zeros((len(means.bin_centers), len(ranges.bin_centers)))
    i_ran_1 = binned_sorted[0][1]
    for en, (i_mea, i_ran, b_mea, b_ran, mea, ran) in enumerate(binned_sorted):
        # print(i_ran, i_ran_1)
        if debug_mode:  # pragma: no cover
            print("---------------------------------")
            print(f"\033[96mEnum - {en}\033[0m")
            print(f"\033[96mID   - Mean: {i_mea}, Range: {i_ran}\033[0m")
            print(f"\033[96mBin  - Mean: {b_mea}, Range: {b_ran}\033[0m")
            print(f"\033[96mReal - Mean: {mea}, Range: {ran}\033[0m")
        # 9) Large cycles identification
        ref_damage = ran**damage_exponent
        approx_damage = b_ran**damage_exponent
        if (
            ref_damage / max_damage > damage_tolerance_for_binning
            and np.abs(ref_damage - approx_damage) / ref_damage
            > damage_tolerance_for_binning
        ) or (i_ran - i_ran_1 > max_consecutive_zeros):
            lg_c.append([mea, ran])
            if debug_mode:  # pragma: no cover
                print(
                    f"\033[92mRea: ({np.round(mea, 3)},"
                    + f" {np.round(ran, 3)}),"
                    + f" Bin: ({np.round(b_mea, 3)},"
                    + f" {np.round(b_ran, 3)}"
                    + ")\033[0m Large Cycle"
                )
            continue
        # 10) Histogram identification
        if debug_mode:  # pragma: no cover
            print(
                f"\033[92mRea: ({np.round(mea, 3)},"
                + f" {np.round(ran, 3)}),"
                + f" Bin: ({np.round(b_mea, 3)},"
                + f" {np.round(b_ran, 3)}"
                + ")\033[0m Histogram Cycle"
            )
        _hist[i_mea][i_ran] += 1
        i_ran_1 = i_ran
    # 11) Histogram binning
    the_hist: List[List[float]] = []
    for hst in _hist:
        the_hist.append(np.trim_zeros(hst, trim="b").tolist())
    return _build_binned_dict(
        ranges, means, lg_c, the_hist, round_decimals, no_mean_stress, out_dct
    )


def rainflow_threshold(ranges: StressRange, means: MeanStress) -> tuple:
    """Removes the rainflow cycles below threshold for a given
    stress range and mean stress.

    Parameters
    ----------
    ranges : StressRange
        The stress range object.
    means : MeanStress
        The mean stress object.

    Returns
    -------
    tuple
        The ranges and means above threshold, as well as small cycles.
    """
    ranges_gt_lb_idx = np.where(ranges.values >= ranges.bin_lb)[0]
    nr_small_cycles = len(ranges.values) - len(ranges_gt_lb_idx)
    means.counts = means.counts[ranges_gt_lb_idx]
    means.values = means.values[ranges_gt_lb_idx]
    ranges.counts = ranges.counts[ranges_gt_lb_idx]
    ranges.values = ranges.values[ranges_gt_lb_idx]
    return ranges, means, nr_small_cycles


def initialise_binned_rainflow_dict(
    ranges: StressRange,
    means: MeanStress,
    nr_small_cycles: int,
    no_mean_stress: bool = False,
) -> dict:
    """Initialises the binned rainflow dictionary.

    Parameters
    ----------
    ranges : StressRange
        The stress range object.
    means : MeanStress
        The mean stress object.
    nr_small_cycles : int
        The number of small cycles.
    no_mean_stress : bool, optional
        Whether to ignore mean stress. Defaults to False.

    Returns
    -------
    dict
        The initialised binned rainflow dictionary.
    """
    out_dct: dict[str, Union[float, list, np.ndarray]] = {}
    out_dct["nr_small_cycles"] = nr_small_cycles
    out_dct["range_bin_lower_bound"] = ranges.bin_lb
    out_dct["range_bin_width"] = ranges.bin_width
    if not no_mean_stress:
        out_dct["mean_bin_lower_bound"] = means.bin_lb
        out_dct["mean_bin_width"] = means.bin_width
    return out_dct


def binned_rainflow(  # pylint: disable=too-many-arguments
    data: Union[np.ndarray, list],
    time: Optional[Union[np.ndarray, list]] = None,
    range_bin_lower_bound: float = 0.2,
    range_bin_width: float = 0.05,
    range_bin_upper_bound: Optional[float] = None,
    mean_bin_lower_bound: Optional[float] = None,
    mean_bin_width: float = 10.0,
    mean_bin_upper_bound: Optional[float] = None,
    damage_tolerance_for_binning: float = 1e-3,
    max_consetutive_zeros: int = 10,
    damage_exponent: float = 5.0,
    round_decimals: int = 4,
    debug_mode: bool = False,
):
    """
    Returns the cycle-count of the input data calculated through the
    :term:`rainflow<Rainflow>` method.

    Parameters
    ----------
    data : np.ndarray
        time series or residuals sequence
    time : Optional[Union[np.ndarray, list]], optional
        sampled times, by default None
    range_bin_lower_bound : float, optional
        lower bound of the range bin, by default 0.2
    range_bin_width : float, optional
        width of the range bin, by default 0.05
    range_bin_upper_bound : Optional[float], optional
        upper bound of the range bin, by default None
    mean_bin_lower_bound : Optional[float], optional
        lower bound of the mean bin, by default None
    mean_bin_width : float, optional
        width of the mean bin, by default 10
    mean_bin_upper_bound : Optional[float], optional
        upper bound of the mean bin, by default None
    damage_tolerance_for_binning : float, optional
        tolerance for the damage when binning large cycles
        separately, by default 1e-3
    damage_exponent : float, optional
        exponent for the damage when binning large cycles separately, by
        default 5.0
    round_decimals : int, optional
        number of decimals to round the output, by default 4
    debug_mode : bool, optional
        if True, print debug messages, by default False
    Returns
    -------
    Union[np.ndarray, tuple]
        - if np.ndarray:
            + rfs : np.ndarray
                rainflow
                [ampl ampl_mean nr_of_cycle cycle_begin_time cycle_period_time]
        - if tuple:
            + rfs : np.ndarray
                rainflow
                [ampl ampl_mean nr_of_cycle cycle_begin_time cycle_period_time]
            + data[res_tp] : numpy.ndarray
                the residuals signal
            + res_tp : numpy.ndarray
                the indices of the residuals signal
            + time[res_tp] : numpy.ndarray
                the time signal for residuals

    Raises
    ------
    TypeError
        data shall be numpy.ndarray or list

    See also
    --------
    calc_rainflow
    rainflow_binner
    """
    rfs, res_seq, _ = calc_rainflow(
        data=data,
        time=time,
        extended_output=True,
    )

    # return rfs, data[res_tp], res_tp, time[res_tp]

    means = MeanStress(
        _counts=rfs[:, 2],
        _values=rfs[:, 1],
        bin_width=mean_bin_width,
        _bin_lb=mean_bin_lower_bound,
        _bin_ub=mean_bin_upper_bound,
    )
    ranges = StressRange(
        _counts=rfs[:, 2],
        _values=2 * rfs[:, 0],
        bin_width=range_bin_width,
        _bin_lb=range_bin_lower_bound,
        _bin_ub=range_bin_upper_bound,
    )
    out_dct = rainflow_binner(
        ranges=ranges,
        _means=means,
        damage_tolerance_for_binning=damage_tolerance_for_binning,
        max_consecutive_zeros=max_consetutive_zeros,
        damage_exponent=damage_exponent,
        round_decimals=round_decimals,
        debug_mode=debug_mode,
    )

    out_dct["res_sig"] = res_seq
    if time is not None:
        out_dct["as_dur"] = time[-1] - time[0]

    return out_dct


def _build_binned_dict(
    ranges: StressRange,
    means: MeanStress,
    lg_c: list,
    the_hist: List[List[float]],
    round_decimals: int,
    no_mean_stress: bool,
    out_dct: dict,
):
    """
    Builds the output rainflow dictionary in py_fatigue format
    after binning the data.

    Parameters
    ----------
    ranges : StressRange
        The stress range object.
    means : MeanStress
        The mean stress object.
    lg_c : list
        The list of large cycles.
    the_hist : List[List[float]]
        The histogram.
    round_decimals : int
        The number of decimals to round the output.
    no_mean_stress : bool
        If True, the mean stress is not included in the output.
    out_dct : dict
        The output dictionary.

    Returns
    -------
    out_dct : dict
        The output dictionary.
    """

    # 11) Saving output to dictionary
    if no_mean_stress:
        out_dct["hist"] = [
            sum(x) for x in itertools.zip_longest(*the_hist, fillvalue=0)
        ]
        out_dct["lg_c"] = (
            np.asarray(np.around(lg_c, round_decimals))[:, 1]
            if len(lg_c) > 0
            else np.empty(0)
        ).tolist()
        out_dct["res"] = np.around(ranges.half, round_decimals).tolist()
    else:
        out_dct["hist"] = the_hist
        out_dct["lg_c"] = np.around(lg_c, round_decimals)
        out_dct["res"] = np.asarray(
            np.around(list(zip(means.half, ranges.half)), round_decimals)
        ).tolist()
    for key, value in out_dct.items():
        if isinstance(value, np.ndarray):
            out_dct[key] = value.tolist()

    return out_dct



"""The module :mod:`py_fatigue.mean_stress.mean_stress` contains
the main mean stress-related utilities.
"""

from typing import ClassVar

from ..utils import FatigueStress


class MeanStress(FatigueStress):
    """Mean stress class."""

    category: ClassVar[str] = "mean"
